""" :mod: GFAL2_XROOTStorage

    =================

    .. module: python

    :synopsis: XROOT module based on the GFAL2_StorageBase class.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

# from DIRAC
from DIRAC import gLogger
from DIRAC.Resources.Storage.GFAL2_StorageBase import GFAL2_StorageBase
from DIRAC.Core.Utilities.Pfn import pfnparse, pfnunparse
from DIRAC.Core.Security.Locations import getProxyLocation

sLog = gLogger.getSubLogger(__name__)


class GFAL2_XROOTStorage(GFAL2_StorageBase):
    """.. class:: GFAL2_XROOTStorage

    Xroot interface to StorageElement using gfal2
    """

    _INPUT_PROTOCOLS = ["file", "root"]
    _OUTPUT_PROTOCOLS = ["root"]

    PROTOCOL_PARAMETERS = GFAL2_StorageBase.PROTOCOL_PARAMETERS + ["SvcClass"]
    DYNAMIC_OPTIONS = {"SvcClass": "svcClass"}

    def __init__(self, storageName, parameters):
        """c'tor

        :param self: self reference
        :param str storageName: SE name
        :param str protocol: protocol to use
        :param str rootdir: base path for vo files
        :param str host: SE host
        :param int port: port to use to communicate with :host:
        :param str spaceToken: space token
        :param str wspath: location of SRM on :host:
        """
        # # init base class
        super(GFAL2_XROOTStorage, self).__init__(storageName, parameters)
        self.srmSpecificParse = False

        self.log = sLog.getSubLogger(storageName)

        self.pluginName = "GFAL2_XROOT"

        # why is this here ?!
        self.protocolParameters["WSUrl"] = 0
        self.protocolParameters["SpaceToken"] = 0

        # We don't need extended attributes for metadata
        self._defaultExtendedAttributes = None

        # Because some storages are configured to use krb5 auth first
        # we end up in trouble for interactive sessions. This
        # environment variable enforces the use of certificates
        if "XrdSecPROTOCOL" not in os.environ:
            os.environ["XrdSecPROTOCOL"] = "gsi,unix"

        # needed for third party copy
        # emulates the gfal2 behaviour: https://its.cern.ch/jira/browse/DMC-1150
        if "XrdSecGSIDELEGPROXY" not in os.environ:
            os.environ["XrdSecGSIDELEGPROXY"] = "1"

    def __addDoubleSlash(self, res):
        """Utilities to add the double slash between the host(:port) and the path

        :param res: DIRAC return structure which contains an URL if S_OK
        :return: DIRAC structure with corrected URL
        """
        if not res["OK"]:
            return res
        url = res["Value"]
        res = pfnparse(url, srmSpecific=self.srmSpecificParse)
        if not res["OK"]:
            return res
        urlDict = res["Value"]
        urlDict["Path"] = "/" + urlDict["Path"]

        # Now, that's one heck of a disgusting hack
        # xrootd client is a bit faulty when managing
        # the connection cache, and ends up reusing an
        # existing connection for different users (security flaw...)
        # they have fixed it (to some extent starting from xrootd 4.10)
        # (https://github.com/xrootd/xrootd/issues/976)
        # BUT. They still can't consume properly the information when
        # the identity is passed in the url (root://url?gsiusrpxy=/tmp/myproxy)
        # So we apply a trick here which is to specify the proxy filename as a virtual user
        # This has no consequence (developer's own words), but to distinguish between users
        # Another ticket has been opened for that https://github.com/xrootd/xrootd/issues/992

        try:
            proxyLoc = getProxyLocation()
            if proxyLoc:
                # xroot does not support dots in the virtual user
                proxyLoc = os.path.basename(proxyLoc).replace(".", "")
                urlDict["Host"] = "%s@%s" % (proxyLoc, urlDict["Host"])
        except Exception as e:
            self.log.warn("Exception trying to add virtual user in the url", repr(e))

        return pfnunparse(urlDict, srmSpecific=self.srmSpecificParse)

    def getURLBase(self, withWSUrl=False):
        """Overwrite to add the double slash"""
        return self.__addDoubleSlash(super(GFAL2_XROOTStorage, self).getURLBase(withWSUrl=withWSUrl))

    def constructURLFromLFN(self, lfn, withWSUrl=False):
        """Overwrite to add the double slash"""
        return self.__addDoubleSlash(super(GFAL2_XROOTStorage, self).constructURLFromLFN(lfn=lfn, withWSUrl=withWSUrl))

    def getCurrentURL(self, fileName):
        """Overwrite to add the double slash"""
        return self.__addDoubleSlash(super(GFAL2_XROOTStorage, self).getCurrentURL(fileName))



""" The MJF utility calculates the amount of wall clock time
    left for a given batch system slot or VM. This is essential for the
    'Filling Mode' where several jobs may be executed in the same slot.

    Machine Job/Features are used following HSF-TN-2016-02 if available.
    Otherwise values are filled in using the batch system and CS
    information.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import ssl
import time
from six.moves.urllib.request import urlopen

import DIRAC

from DIRAC import gLogger, gConfig

__RCSID__ = "$Id$"


class MJF(object):
    """Machine/Job Features methods"""

    mjfKeys = {
        "MACHINEFEATURES": ["total_cpu", "hs06", "shutdowntime", "grace_secs"],
        "JOBFEATURES": [
            "allocated_cpu",
            "hs06_job",
            "shutdowntime_job",
            "grace_secs_job",
            "jobstart_secs",
            "job_id",
            "wall_limit_secs",
            "cpu_limit_secs",
            "max_rss_bytes",
            "max_swap_bytes",
            "scratch_limit_bytes",
        ],
    }

    #############################################################################
    def __init__(self):
        """Standard constructor"""
        self.log = gLogger.getSubLogger("MJF")

        capath = DIRAC.Core.Security.Locations.getCAsLocation()

        if not capath:
            raise Exception("Unable to find CA files location! Not in /etc/grid-security/certificates/ etc.")

        # Used by urllib when talking to HTTPS web servers
        self.context = ssl.create_default_context(capath=capath)

    def updateConfig(self, pilotStartTime=None):
        """Populate /LocalSite/MACHINEFEATURES and /LocalSite/JOBFEATURES with MJF values
        This is run early in the job to update the configuration file that subsequent DIRAC
        scripts read when they start.
        """

        if pilotStartTime:
            gConfig.setOptionValue("/LocalSite/JOBFEATURES/jobstart_secs", str(pilotStartTime))

        for mORj in ["MACHINEFEATURES", "JOBFEATURES"]:
            for key in self.mjfKeys[mORj]:
                value = self.__fetchMachineJobFeature(mORj, key)

                if value is not None:
                    gConfig.setOptionValue("/LocalSite/%s/%s" % (mORj, key), value)

    def getMachineFeature(self, key):
        """Returns MACHINEFEATURES/key value saved in /LocalSite configuration by
        updateConfigFile() unless MACHINEFEATURES/shutdowntime when we try to fetch
        from the source URL itself again in case it changes.
        """
        if key == "shutdowntime":
            value = self.__fetchMachineJobFeature("MACHINEFEATURES", "shutdowntime")
            # If unable to fetch shutdowntime, go back to any value in /LocalSite
            # in case HTTP(S) server is down
            if value is not None:
                return value

        return gConfig.getValue("/LocalSite/MACHINEFEATURES/" + key, None)

    def getIntMachineFeature(self, key):
        """Returns MACHINEFEATURES/key as an int or None if not an int or not present"""
        value = self.getMachineFeature(key)

        try:
            return int(value)
        except ValueError:
            return None

    def getJobFeature(self, key):
        """Returns JOBFEATURES/key value saved in /LocalSite configuration by
        updateConfigFile() unless JOBFEATURES/shutdowntime_job when we try to fetch
        from the source URL itself again in case it changes.
        """
        if key == "shutdowntime_job":
            value = self.__fetchMachineJobFeature("JOBFEATURES", "shutdowntime_job")
            # If unable to fetch shutdowntime_job, go back to any value in /LocalSite
            # in case HTTP(S) server is down
            if value is not None:
                return value

        return gConfig.getValue("/LocalSite/JOBFEATURES/" + key, None)

    def getIntJobFeature(self, key):
        """Returns JOBFEATURES/key as an int or None if not an int or not present"""
        value = self.getJobFeature(key)

        try:
            return int(value)
        except ValueError:
            return None

    def __fetchMachineJobFeature(self, mORj, key):
        """Returns raw MJF value for a given key, perhaps by HTTP(S), perhaps from a local file
        mORj must be MACHINEFEATURES or JOBFEATURES
        If the value cannot be found, then return None. There are many legitimate ways for
        a site not to provide some MJF values so we don't log errors, failures etc.
        """
        if mORj != "MACHINEFEATURES" and mORj != "JOBFEATURES":
            raise Exception("Must request MACHINEFEATURES or JOBFEATURES")

        if mORj not in os.environ:
            return None

        url = os.environ[mORj] + "/" + key

        # Simple if a file
        if url[0] == "/":
            try:
                with open(url, "r") as fd:
                    return fd.read().strip()
            except Exception:
                return None

        # Otherwise make sure it's an HTTP(S) URL
        if not url.startswith("http://") and not url.startswith("https://"):
            return None

        # We could have used urlopen() for local files too, but we also
        # need to check HTTP return code in case we get an HTML error page
        # instead of a true key value.
        try:
            mjfUrl = urlopen(url=url, context=self.context)
            # HTTP return codes other than 2xx mean failure
            if int(mjfUrl.getcode() / 100) != 2:
                return None
            return mjfUrl.read().strip()
        except Exception:
            return None
        finally:
            try:
                mjfUrl.close()
            except UnboundLocalError:
                pass

    def getWallClockSecondsLeft(self):
        """Returns the number of seconds until either the wall clock limit
        or the shutdowntime(_job) is reached.
        """

        now = int(time.time())
        secondsLeft = None
        jobstartSecs = self.getIntJobFeature("jobstart_secs")
        wallLimitSecs = self.getIntJobFeature("wall_limit_secs")
        shutdowntimeJob = self.getIntJobFeature("shutdowntime_job")
        shutdowntime = self.getIntMachineFeature("shutdowntime")

        # look for local shutdown file
        try:
            with open("/var/run/shutdown_time", "r") as fd:
                shutdowntimeLocal = int(fd.read().strip())
        except (IOError, ValueError):
            shutdowntimeLocal = None

        if jobstartSecs is not None and wallLimitSecs is not None:
            secondsLeft = jobstartSecs + wallLimitSecs - now

        if shutdowntimeJob is not None:
            if secondsLeft is None:
                secondsLeft = shutdowntimeJob - now
            elif shutdowntimeJob - now < secondsLeft:
                secondsLeft = shutdowntimeJob - now

        if shutdowntime is not None:
            if secondsLeft is None:
                secondsLeft = shutdowntime - now
            elif shutdowntime - now < secondsLeft:
                secondsLeft = shutdowntime - now

        if shutdowntimeLocal is not None:
            if secondsLeft is None:
                secondsLeft = shutdowntimeLocal - now
            elif shutdowntimeLocal - now < secondsLeft:
                secondsLeft = shutdowntimeLocal - now

        # Wall Clock time left or None if unknown
        return secondsLeft



"""
Class for management of MQ communication
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Utilities.DErrno import EMQUKN

__RCSID__ = "$Id$"


def createMQConnector(parameters=None):
    """Function creates and returns the MQConnector object based.

    Args:
      parameters(dict): set of parameters for the MQConnector constructor,
        it should also contain pair 'MQType':mqType, where
        mqType is a string used as a prefix for the specialized MQConnector
        class.
    Returns:
      S_OK or S_ERROR: with loaded specialized class of MQConnector.
    """
    mqType = parameters.get("MQType", None)
    result = getMQConnectorClass(mqType=mqType)
    if not result["OK"]:
        gLogger.error("Failed to getMQConnectorClass:", "%s" % (result["Message"]))
        return result
    ceClass = result["Value"]
    try:
        mqConnector = ceClass(parameters)
        if not result["OK"]:
            return result
    except Exception as exc:  # pylint: disable=broad-except
        gLogger.exception("Could not instantiate MQConnector object", lExcInfo=exc)
        return S_ERROR(EMQUKN, "")
    return S_OK(mqConnector)


def getMQConnectorClass(mqType):
    """Function loads the specialized MQConnector class based on mqType.
        It is assumed that MQConnector has a name in the format mqTypeMQConnector
        e.g. if StompMQConnector.

    Args:
      mqType(str): prefix of specialized class name e.g. Stomp.
    Returns:
      S_OK or S_ERROR: with loaded specialized class of MQConnector.
    """
    subClassName = mqType + "MQConnector"
    result = ObjectLoader().loadObject("Resources.MessageQueue.%s" % subClassName)
    if not result["OK"]:
        gLogger.error("Failed to load object", "%s: %s" % (subClassName, result["Message"]))
    return result


class MQConnectionError(Exception):
    """specialized exception"""

    pass


class MQConnector(object):
    """
    Class for management of message queue connections
    """

    def __init__(self, parameters=None):
        """Standard constructor"""
        if not parameters:
            parameters = {}
        self.parameters = parameters

    def setupConnection(self, parameters=None):
        """
        :param dict parameters: dictionary with additional MQ parameters if any
        :return: S_OK/S_ERROR
        """
        raise NotImplementedError("This method should be implemented by child class")

    def put(self, message, parameters=None):
        """Send message to a MQ server

        :param message: any json encodable structure
        :return: S_OK/S_ERROR
        """
        raise NotImplementedError("This method should be implemented by child class")

    def connect(self, parameters=None):
        """
        :param dict parameters: dictionary with additional parameters if any
        :return: S_OK/S_ERROR
        """
        raise NotImplementedError("This method should be implemented by child class")

    def disconnect(self, parameters=None):
        """
        Disconnects from the message queue server

        :param dict parameters: dictionary with additional parameters if any
        :return: S_OK/S_ERROR
        """
        raise NotImplementedError("This method should be implemented by child class")

    def subscribe(self, parameters=None):
        """
        Subscribes to the message queue server

        :param dict parameters: dictionary with additional parameters if any
        :return: S_OK/S_ERROR
        """

        raise NotImplementedError("This method should be implemented by child class")

    def unsubscribe(self, parameters=None):
        """
        Subscribes to the message queue server

        :param dict parameters: dictionary with additional parameters if any
        :return: S_OK/S_ERROR
        """
        raise NotImplementedError("This method should be implemented by child class")



#!/usr/bin/env python
########################################################################
# File :    dirac-dms-pfn-accessURL
# Author  : Stuart Paterson
########################################################################
"""
Retrieve an access URL for a PFN given a valid DIRAC SE
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("PFN:      Physical File Name or file containing PFNs")
    Script.registerArgument("SE:       Valid DIRAC SE")
    _, args = Script.parseCommandLine(ignoreErrors=True)

    if len(args) > 2:
        print("Only one PFN SE pair will be considered")

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    exitCode = 0

    pfn = args[0]
    seName = args[1]
    try:
        with open(pfn, "r") as f:
            pfns = f.read().splitlines()
    except Exception:
        pfns = [pfn]

    for pfn in pfns:
        result = dirac.getPhysicalFileAccessURL(pfn, seName, printOutput=True)
        if not result["OK"]:
            print("ERROR: ", result["Message"])
            exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.AccountingSystem.Client.Types.WMSHistory import WMSHistory
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter


class WMSHistoryPlotter(BaseReporter):

    _typeName = "WMSHistory"
    _typeKeyFields = [dF[0] for dF in WMSHistory().definitionKeyFields]

    def _translateGrouping(self, grouping):
        if grouping == "Country":
            sqlRepr = 'upper( substring( %s, locate( ".", %s, length( %s ) - 4 ) + 1 ) )'
            return (sqlRepr, ["Site", "Site", "Site"], sqlRepr)
        elif grouping == "Grid":
            return ('substring_index( %s, ".", 1 )', ["Site"])
        else:
            return ("%s", [grouping])

    def _reportNumberOfJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Jobs", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotNumberOfJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Jobs by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "skipEdgeColor": True,
            "ylabel": "jobs",
        }
        plotInfo["data"] = self._fillWithZero(
            plotInfo["granularity"], reportRequest["startTime"], reportRequest["endTime"], plotInfo["data"]
        )
        return self._generateStackedLinePlot(filename, plotInfo["data"], metadata)

    def _reportNumberOfReschedules(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Reschedules", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {"convertToGranularity": "average", "checkNone": True},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotNumberOfReschedules(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Reschedules by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "skipEdgeColor": True,
            "ylabel": "reschedules",
        }
        plotInfo["data"] = self._fillWithZero(
            plotInfo["granularity"], reportRequest["startTime"], reportRequest["endTime"], plotInfo["data"]
        )
        return self._generateStackedLinePlot(filename, plotInfo["data"], metadata)

    def _reportAverageNumberOfJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", SUM(%s/%s)",
            reportRequest["groupingFields"][1] + ["Jobs", "entriesInBucket"],
        )
        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        bins = self._getBins(self._typeName, reportRequest["startTime"], reportRequest["endTime"])
        numBins = len(bins)
        for key in dataDict:
            dataDict[key] = float(dataDict[key] / numBins)
        return S_OK({"data": dataDict})

    def _plotAverageNumberOfJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Average Number of Jobs by %s" % reportRequest["grouping"],
            "ylabel": "Jobs",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)



#!/usr/bin/env python
########################################################################
# File : dirac-wms-job-delete
# Author : Stuart Paterson
########################################################################
"""
Delete DIRAC job from WMS, if running it will be killed

Example:
  $ dirac-wms-job-delete 12
  Deleted job 12
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os.path

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("f:", "File=", "Get output for jobs with IDs from the file")
    Script.registerSwitch("g:", "JobGroup=", "Get output for jobs in the given group")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"], mandatory=False)
    sws, args = Script.parseCommandLine(ignoreErrors=True)

    import DIRAC
    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments
    from DIRAC.Core.Utilities.Time import toString, date, day

    dirac = Dirac()

    jobs = []
    for sw, value in sws:
        if sw.lower() in ("f", "file"):
            if os.path.exists(value):
                jFile = open(value)
                jobs += jFile.read().split()
                jFile.close()
        elif sw.lower() in ("g", "jobgroup"):
            group = value
            jobDate = toString(date() - 30 * day)
            result = dirac.selectJobs(jobGroup=value, date=jobDate)
            if not result["OK"]:
                if "No jobs selected" not in result["Message"]:
                    print("Error:", result["Message"])
                    DIRAC.exit(-1)
            else:
                jobs += result["Value"]

    for arg in parseArguments(args):
        jobs.append(arg)

    if not jobs:
        print("Warning: no jobs selected")
        Script.showHelp()
        DIRAC.exit(0)

    result = dirac.deleteJob(jobs)
    if result["OK"]:
        print("Deleted jobs %s" % ",".join([str(j) for j in result["Value"]]))
        exitCode = 0
    else:
        print(result["Message"])
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



########################################################################
# File: RequestTask.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/13 12:42:45
########################################################################

""" :mod: RequestTask

    =================

    .. module: RequestTask

    :synopsis: request processing task

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    request processing task to be used inside ProcessTask created in RequesteExecutingAgent
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# #
# @file RequestTask.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/13 12:42:54
# @brief Definition of RequestTask class.
# # imports
import os
import time
import datetime

# # from DIRAC
from DIRAC import gLogger, S_OK, S_ERROR, gConfig
from DIRAC.Core.Utilities import DErrno
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.Core.Utilities import Time, Network
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient

########################################################################


class RequestTask(object):
    """
    .. class:: RequestTask

    request's processing task
    """

    def __init__(
        self, requestJSON, handlersDict, csPath, agentName, standalone=False, requestClient=None, rmsMonitoring=False
    ):
        """c'tor

        :param self: self reference
        :param str requestJSON: request serialized to JSON
        :param dict opHandlers: operation handlers
        """
        self.request = Request(requestJSON)
        # # csPath
        self.csPath = csPath
        # # agent name
        self.agentName = agentName
        # # standalone flag
        self.standalone = standalone
        # # handlers dict
        self.handlersDict = handlersDict
        # # handlers class def
        self.handlers = {}
        # # own sublogger
        self.log = gLogger.getSubLogger("pid_%s/%s" % (os.getpid(), self.request.RequestName))
        # # get shifters info
        self.__managersDict = {}
        shifterProxies = self.__setupManagerProxies()
        if not shifterProxies["OK"]:
            self.log.error("Cannot setup shifter proxies", shifterProxies["Message"])

        #  This flag which is set and sent from the RequestExecutingAgent and is False by default.
        self.rmsMonitoring = rmsMonitoring

        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # initialize gMonitor
            gMonitor.setComponentType(gMonitor.COMPONENT_AGENT)
            gMonitor.setComponentName(self.agentName)
            gMonitor.initialize()

            # # own gMonitor activities
            gMonitor.registerActivity(
                "RequestAtt", "Requests processed", "RequestExecutingAgent", "Requests/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RequestFail", "Requests failed", "RequestExecutingAgent", "Requests/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RequestOK", "Requests done", "RequestExecutingAgent", "Requests/min", gMonitor.OP_SUM
            )

        if requestClient is None:
            self.requestClient = ReqClient()
        else:
            self.requestClient = requestClient

    def __setupManagerProxies(self):
        """setup grid proxy for all defined managers"""
        oHelper = Operations()
        shifters = oHelper.getSections("Shifter")
        if not shifters["OK"]:
            return shifters
        shifters = shifters["Value"]
        for shifter in shifters:
            shifterDict = oHelper.getOptionsDict("Shifter/%s" % shifter)
            if not shifterDict["OK"]:
                self.log.error("Cannot get options dict for shifter", "%s: %s" % (shifter, shifterDict["Message"]))
                continue
            userName = shifterDict["Value"].get("User", "")
            userGroup = shifterDict["Value"].get("Group", "")

            userDN = Registry.getDNForUsername(userName)
            if not userDN["OK"]:
                self.log.error("Cannot get DN For Username", "%s: %s" % (userName, userDN["Message"]))
                continue
            userDN = userDN["Value"][0]
            vomsAttr = Registry.getVOMSAttributeForGroup(userGroup)
            if vomsAttr:
                self.log.debug(
                    "getting VOMS [%s] proxy for shifter %s@%s (%s)" % (vomsAttr, userName, userGroup, userDN)
                )
                getProxy = gProxyManager.downloadVOMSProxyToFile(
                    userDN, userGroup, requiredTimeLeft=1200, cacheTime=4 * 43200
                )
            else:
                self.log.debug("getting proxy for shifter %s@%s (%s)" % (userName, userGroup, userDN))
                getProxy = gProxyManager.downloadProxyToFile(
                    userDN, userGroup, requiredTimeLeft=1200, cacheTime=4 * 43200
                )
            if not getProxy["OK"]:
                return S_ERROR("unable to setup shifter proxy for %s: %s" % (shifter, getProxy["Message"]))
            chain = getProxy["chain"]
            fileName = getProxy["Value"]
            self.log.debug("got %s: %s %s" % (shifter, userName, userGroup))
            self.__managersDict[shifter] = {
                "ShifterDN": userDN,
                "ShifterName": userName,
                "ShifterGroup": userGroup,
                "Chain": chain,
                "ProxyFile": fileName,
            }
        return S_OK()

    def setupProxy(self):
        """download and dump request owner proxy to file and env

        :return: S_OK with name of newly created owner proxy file and shifter name if any
        """
        self.__managersDict = {}
        shifterProxies = self.__setupManagerProxies()
        if not shifterProxies["OK"]:
            self.log.error(shifterProxies["Message"])

        ownerDN = self.request.OwnerDN
        ownerGroup = self.request.OwnerGroup
        isShifter = []
        for shifter, creds in self.__managersDict.items():
            if creds["ShifterDN"] == ownerDN and creds["ShifterGroup"] == ownerGroup:
                isShifter.append(shifter)
        if isShifter:
            proxyFile = self.__managersDict[isShifter[0]]["ProxyFile"]
            os.environ["X509_USER_PROXY"] = proxyFile
            return S_OK({"Shifter": isShifter, "ProxyFile": proxyFile})

        # # if we're here owner is not a shifter at all
        ownerProxyFile = gProxyManager.downloadVOMSProxyToFile(ownerDN, ownerGroup)
        if not ownerProxyFile["OK"] or not ownerProxyFile["Value"]:
            reason = ownerProxyFile.get("Message", "No valid proxy found in ProxyManager.")
            return S_ERROR("Change proxy error for '%s'@'%s': %s" % (ownerDN, ownerGroup, reason))

        ownerProxyFile = ownerProxyFile["Value"]
        os.environ["X509_USER_PROXY"] = ownerProxyFile
        return S_OK({"Shifter": isShifter, "ProxyFile": ownerProxyFile})

    @staticmethod
    def getPluginName(pluginPath):
        """return plugin name"""
        if not pluginPath:
            return ""
        if "/" in pluginPath:
            pluginPath = ".".join([chunk for chunk in pluginPath.split("/") if chunk])
        return pluginPath.split(".")[-1]

    def loadHandler(self, pluginPath):
        """Create an instance of requested plugin class, loading and importing it when needed.
        This function could raise ImportError when plugin cannot be find or TypeError when
        loaded class object isn't inherited from BaseOperation class.

        :param str pluginName: dotted path to plugin, specified as in import statement, i.e.
            "DIRAC.CheesShopSystem.private.Cheddar" or alternatively in 'normal' path format
            "DIRAC/CheesShopSystem/private/Cheddar"

        :return: object instance

        This function try to load and instantiate an object from given path. It is assumed that:

          * `pluginPath` is pointing to module directory "importable" by python interpreter, i.e.: it's
            package's top level directory is in $PYTHONPATH env variable,
          * the module should consist a class definition following module name,
          *  the class itself is inherited from DIRAC.RequestManagementSystem.private.BaseOperation.BaseOperation

        If above conditions aren't meet, function is throwing exceptions:

        :raises ImportError: when class cannot be imported
        :raises TypeError: when class isn't inherited from OperationHandlerBase
        """
        if "/" in pluginPath:
            pluginPath = ".".join([chunk for chunk in pluginPath.split("/") if chunk])
        pluginName = pluginPath.split(".")[-1]
        if pluginName not in globals():
            mod = __import__(pluginPath, globals(), fromlist=[pluginName])
            pluginClassObj = getattr(mod, pluginName)
        else:
            pluginClassObj = globals()[pluginName]
        if not issubclass(pluginClassObj, OperationHandlerBase):
            raise TypeError("operation handler '%s' isn't inherited from OperationHandlerBase class" % pluginName)

        if not self.rmsMonitoring:
            for key, status in (("Att", "Attempted"), ("OK", "Successful"), ("Fail", "Failed")):
                gMonitor.registerActivity(
                    "%s%s" % (pluginName, key),
                    "%s operations %s" % (pluginName, status),
                    "RequestExecutingAgent",
                    "Operations/min",
                    gMonitor.OP_SUM,
                )
        # # return an instance
        return pluginClassObj

    def getHandler(self, operation):
        """return instance of a handler for a given operation type on demand
            all created handlers are kept in self.handlers dict for further use

        :param ~Operation.Operation operation: Operation instance
        """
        if operation.Type not in self.handlersDict:
            return S_ERROR("handler for operation '%s' not set" % operation.Type)
        handler = self.handlers.get(operation.Type, None)
        if not handler:
            try:
                handlerCls = self.loadHandler(self.handlersDict[operation.Type])
                self.handlers[operation.Type] = handlerCls(
                    csPath="%s/OperationHandlers/%s" % (self.csPath, operation.Type)
                )
                handler = self.handlers[operation.Type]
            except (ImportError, TypeError) as error:
                self.log.exception("Error getting Handler", "%s" % error, lException=error)
                return S_ERROR(str(error))
        # # set operation for this handler
        handler.setOperation(operation)
        # # and return
        return S_OK(handler)

    def updateRequest(self):
        """put back request to the RequestDB"""
        updateRequest = self.requestClient.putRequest(self.request, useFailoverProxy=False, retryMainService=2)
        if not updateRequest["OK"]:
            self.log.error("Cannot updateRequest", updateRequest["Message"])
        return updateRequest

    def __call__(self):
        """request processing"""

        self.log.debug("about to execute request")
        if not self.rmsMonitoring:
            gMonitor.addMark("RequestAtt", 1)

        # # setup proxy for request owner
        setupProxy = self.setupProxy()
        if not setupProxy["OK"]:
            userSuspended = "User is currently suspended"
            self.request.Error = setupProxy["Message"]
            # In case the user does not have proxy
            if DErrno.cmpError(setupProxy, DErrno.EPROXYFIND):
                self.log.error("Error setting proxy. Request set to Failed:", setupProxy["Message"])
                # If user is no longer registered, fail the request
                for operation in self.request:
                    for opFile in operation:
                        opFile.Status = "Failed"
                    operation.Status = "Failed"
            elif userSuspended in setupProxy["Message"]:
                # If user is suspended, wait for a long time
                self.request.delayNextExecution(6 * 60)
                self.request.Error = userSuspended
                self.log.error("Error setting proxy: " + userSuspended, self.request.OwnerDN)
            else:
                self.log.error("Error setting proxy", setupProxy["Message"])
            return S_OK(self.request)
        shifter = setupProxy["Value"]["Shifter"]

        error = None

        while self.request.Status == "Waiting":

            # # get waiting operation
            operation = self.request.getWaiting()
            if not operation["OK"]:
                self.log.error("Cannot get waiting operation", operation["Message"])
                return operation
            operation = operation["Value"]
            self.log.info("executing operation", "%s" % operation.Type)

            # # and handler for it
            handler = self.getHandler(operation)
            if not handler["OK"]:
                self.log.error("Unable to process operation", "%s: %s" % (operation.Type, handler["Message"]))
                # gMonitor.addMark( "%s%s" % ( operation.Type, "Fail" ), 1 )
                operation.Error = handler["Message"]
                break

            handler = handler["Value"]
            # # set shifters list in the handler
            handler.shifter = shifter
            # set rmsMonitoring flag for the RequestOperation
            handler.rmsMonitoring = self.rmsMonitoring
            # # and execute
            pluginName = self.getPluginName(self.handlersDict.get(operation.Type))
            if self.standalone:
                useServerCertificate = gConfig.useServerCertificate()
            else:
                # Always use server certificates if executed within an agent
                useServerCertificate = True
            try:
                if pluginName:
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(
                            {
                                "timestamp": int(Time.toEpoch()),
                                "host": Network.getFQDN(),
                                "objectType": "Operation",
                                "operationType": pluginName,
                                "objectID": operation.OperationID,
                                "parentID": operation.RequestID,
                                "status": "Attempted",
                                "nbObject": 1,
                            }
                        )
                    else:
                        gMonitor.addMark("%s%s" % (pluginName, "Att"), 1)
                # Always use request owner proxy
                if useServerCertificate:
                    gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")
                exe = handler()
                if useServerCertificate:
                    gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
                if not exe["OK"]:
                    self.log.error("unable to process operation", "%s: %s" % (operation.Type, exe["Message"]))
                    if pluginName:
                        if self.rmsMonitoring:
                            self.rmsMonitoringReporter.addRecord(
                                {
                                    "timestamp": int(Time.toEpoch()),
                                    "host": Network.getFQDN(),
                                    "objectType": "Operation",
                                    "operationType": pluginName,
                                    "objectID": operation.OperationID,
                                    "parentID": operation.RequestID,
                                    "status": "Failed",
                                    "nbObject": 1,
                                }
                            )
                        else:
                            gMonitor.addMark("%s%s" % (pluginName, "Fail"), 1)
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(
                            {
                                "timestamp": int(Time.toEpoch()),
                                "host": Network.getFQDN(),
                                "objectType": "Request",
                                "objectID": operation.RequestID,
                                "status": "Failed",
                                "nbObject": 1,
                            }
                        )
                    else:
                        gMonitor.addMark("RequestFail", 1)

                    if self.request.JobID:
                        # Check if the job exists
                        monitorServer = JobMonitoringClient(useCertificates=True)
                        res = monitorServer.getJobSummary(int(self.request.JobID))
                        if not res["OK"]:
                            self.log.error("RequestTask: Failed to get job status", "%d" % self.request.JobID)
                        elif not res["Value"]:
                            self.log.warn(
                                "RequestTask: job does not exist (anymore): failed request",
                                "JobID: %d" % self.request.JobID,
                            )
                            for opFile in operation:
                                opFile.Status = "Failed"
                            if operation.Status != "Failed":
                                operation.Status = "Failed"
                            self.request.Error = "Job no longer exists"
            except Exception as error:
                self.log.exception("hit by exception:", "%s" % error)
                if pluginName:
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(
                            {
                                "timestamp": int(Time.toEpoch()),
                                "host": Network.getFQDN(),
                                "objectType": "Operation",
                                "operationType": pluginName,
                                "objectID": operation.OperationID,
                                "parentID": operation.RequestID,
                                "status": "Failed",
                                "nbObject": 1,
                            }
                        )
                    else:
                        gMonitor.addMark("%s%s" % (pluginName, "Fail"), 1)
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(
                        {
                            "timestamp": int(Time.toEpoch()),
                            "host": Network.getFQDN(),
                            "objectType": "Request",
                            "objectID": operation.RequestID,
                            "status": "Failed",
                            "nbObject": 1,
                        }
                    )
                else:
                    gMonitor.addMark("RequestFail", 1)

                if useServerCertificate:
                    gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
                break

            # # operation status check
            if operation.Status == "Done" and pluginName:
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(
                        {
                            "timestamp": int(Time.toEpoch()),
                            "host": Network.getFQDN(),
                            "objectType": "Operation",
                            "operationType": pluginName,
                            "objectID": operation.OperationID,
                            "parentID": operation.RequestID,
                            "status": "Successful",
                            "nbObject": 1,
                        }
                    )
                else:
                    gMonitor.addMark("%s%s" % (pluginName, "OK"), 1)
            elif operation.Status == "Failed" and pluginName:
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(
                        {
                            "timestamp": int(Time.toEpoch()),
                            "host": Network.getFQDN(),
                            "objectType": "Operation",
                            "operationType": pluginName,
                            "objectID": operation.OperationID,
                            "parentID": operation.RequestID,
                            "status": "Failed",
                            "nbObject": 1,
                        }
                    )
                else:
                    gMonitor.addMark("%s%s" % (pluginName, "Fail"), 1)
            elif operation.Status in ("Waiting", "Scheduled"):
                # # no update for waiting or all files scheduled
                break

        if not self.rmsMonitoring:
            gMonitor.flush()

        if error:
            return S_ERROR(error)

        # # request done?
        if self.request.Status == "Done":
            # # update request to the RequestDB
            self.log.info("Updating request status:", "%s" % self.request.Status)
            update = self.updateRequest()
            if not update["OK"]:
                self.log.error("Cannot update request status", update["Message"])
                return update
            self.log.info("request is done", "%s" % self.request.RequestName)
            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(
                    {
                        "timestamp": int(Time.toEpoch()),
                        "host": Network.getFQDN(),
                        "objectType": "Request",
                        "objectID": getattr(self.request, "RequestID", 0),
                        "status": "Successful",
                        "nbObject": 1,
                    }
                )
            else:
                gMonitor.addMark("RequestOK", 1)
            # # and there is a job waiting for it? finalize!
            if self.request.JobID:
                attempts = 0
                while True:
                    finalizeRequest = self.requestClient.finalizeRequest(
                        self.request.RequestID, self.request.JobID  # pylint: disable=no-member
                    )
                    if not finalizeRequest["OK"]:
                        if not attempts:
                            self.log.error(
                                "unable to finalize request, will retry",
                                "ReqName %s:%s" % (self.request.RequestName, finalizeRequest["Message"]),
                            )
                        self.log.debug("Waiting 10 seconds")
                        attempts += 1
                        if attempts == 10:
                            self.log.error("Giving up finalize request")
                            return S_ERROR("Could not finalize request")

                        time.sleep(10)

                    else:
                        self.log.info(
                            "request is finalized",
                            "ReqName %s %s"
                            % (self.request.RequestName, (" after %d attempts" % attempts) if attempts else ""),
                        )
                        break

        # Commit all the data to the ES Backend
        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()
        # Request will be updated by the callBack method
        self.log.verbose("RequestTasks exiting", "request %s" % self.request.Status)
        return S_OK(self.request)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


class ColorGenerator:
    def __init__(self):
        self.lUsedColors = []

    def __equalColors(self, c1, c2):
        return c1[0] == c2[0] and c1[1] == c2[1] and c1[2] == c2[2]

    def __equalColorInList(self, c1, cl2):
        return any(self.__equalColors(c1, c2) for c2 in cl2)

    def __toHex(self, n):
        return hex(int(n))[2:].upper().rjust(2, "0")

    def reset(self):
        self.lUsedColors = []

    def __generateColor(self):
        iNumColorsUsed = len(self.lUsedColors)
        if iNumColorsUsed == 0:
            self.lUsedColors.append((0, 0, 255))
        elif iNumColorsUsed == 1:
            self.lUsedColors.append((0, 255, 0))
        elif iNumColorsUsed == 2:
            self.lUsedColors.append((255, 0, 0))
        elif iNumColorsUsed == 3:
            self.lUsedColors.append((0, 255, 255))
        elif iNumColorsUsed == 4:
            self.lUsedColors.append((255, 0, 255))
        elif iNumColorsUsed == 5:
            self.lUsedColors.append((255, 255, 0))
        elif len(self.lUsedColors) < 128:
            # This has terrible performance is the number of used colours gets too high
            # At that point the color stops making much sense so allow it to loop
            im = 0
            iM = 1
            rC = self.lUsedColors[0]
            while self.__equalColorInList(rC, self.lUsedColors):
                stC1 = self.lUsedColors[im]
                stC2 = self.lUsedColors[iM]
                rC = ((stC1[0] + stC2[0]) / 2, (stC1[1] + stC2[1]) / 2, (stC1[2] + stC2[2]) / 2)
                if self.__equalColorInList(rC, self.lUsedColors):
                    im += 1
                    if im == iM:
                        iM += 1
                        im = 0
            self.lUsedColors.append(rC)
        return self.lUsedColors[iNumColorsUsed % 128]

    def getFloatColor(self):
        fC = self.__generateColor()
        return (float(fC[0]) / 255, float(fC[1]) / 255, float(fC[2]) / 255)

    def getHexColor(self):
        fC = self.__generateColor()
        return "%s%s%s" % (self.__toHex(fC[0]), self.__toHex(fC[1]), self.__toHex(fC[2]))



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-get-queue-cpu-time.py
# Author :  Federico Stagni
########################################################################
"""
Report CPU length of queue, in seconds
This script is used by the dirac-pilot script to set the CPUTime left, which is a limit for the matching
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("C:", "CPUNormalizationFactor=", "CPUNormalizationFactor, in case it is known")
    Script.parseCommandLine(ignoreErrors=True)

    CPUNormalizationFactor = 0.0
    for unprocSw in Script.getUnprocessedSwitches():
        if unprocSw[0] in ("C", "CPUNormalizationFactor"):
            CPUNormalizationFactor = float(unprocSw[1])

    from DIRAC.WorkloadManagementSystem.Client.CPUNormalization import getCPUTime

    cpuTime = getCPUTime(CPUNormalizationFactor)
    # I hate this kind of output... PhC
    print("CPU time left determined as", cpuTime)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
""" This template will become the job wrapper that's actually executed.

    The JobWrapperTemplate is completed and invoked by the jobAgent and uses functionalities from JobWrapper module.
    It has to be an executable.

    The JobWrapperTemplate will reschedule the job according to certain criteria:
    - the working directory could not be created
    - the jobWrapper initialization phase failed
    - the inputSandbox download failed
    - the resolution of the inpt data failed
    - the JobWrapper ended with the status DErrno.EWMSRESC
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import json
import ast
import os
import errno
import time
import signal

sitePython = "@SITEPYTHON@"
if sitePython:
    sys.path.insert(0, "@SITEPYTHON@")

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

Script.parseCommandLine()

from DIRAC import gLogger
from DIRAC.Core.Utilities import DErrno

from DIRAC.WorkloadManagementSystem.JobWrapper.JobWrapper import JobWrapper, rescheduleFailedJob
from DIRAC.WorkloadManagementSystem.Client.JobReport import JobReport
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus


gJobReport = None


os.umask(0o22)


class JobWrapperError(Exception):
    """Custom exception for handling JobWrapper "genuine" errors"""

    def __init__(self, value):
        self.value = value
        super(JobWrapperError, self).__init__()

    def __str__(self):
        return str(self.value)


def killJobWrapper(job):
    """Function that stops and ultimately kills the JobWrapper"""
    # Giving the JobWrapper some time to complete possible tasks, then trying to kill the process
    time.sleep(60)
    os.kill(job.currentPID, signal.SIGTERM)
    # wait for half a minute and if worker is still alive use REAL silencer
    time.sleep(30)
    # now you're dead
    os.kill(job.currentPID, signal.SIGKILL)
    return 1


def sendJobAccounting(job, status, minorStatus):
    """safe sending job accounting (always catching exceptions)"""
    try:
        job.sendJobAccounting(status, minorStatus)
    except Exception as exc:  # pylint: disable=broad-except
        gLogger.exception(
            "JobWrapper failed sending job accounting for [status:minorStatus] [%s:%s]" % (status, minorStatus),
            lException=exc,
        )


def execute(arguments):
    """The only real function executed here"""

    global gJobReport

    jobID = arguments["Job"].get("JobID", 0)
    os.environ["JOBID"] = str(jobID)
    jobID = int(jobID)

    if "WorkingDirectory" in arguments:
        wdir = os.path.expandvars(arguments["WorkingDirectory"])
        if os.path.isdir(wdir):
            os.chdir(wdir)
        else:
            try:
                os.makedirs(wdir)  # this will raise an exception if wdir already exists (which is ~OK)
                if os.path.isdir(wdir):
                    os.chdir(wdir)
            except OSError as osError:
                if osError.errno == errno.EEXIST and os.path.isdir(wdir):
                    gLogger.exception("JobWrapperTemplate found that the working directory already exists")
                    rescheduleResult = rescheduleFailedJob(jobID, "Working Directory already exists")
                else:
                    gLogger.exception("JobWrapperTemplate could not create working directory")
                    rescheduleResult = rescheduleFailedJob(jobID, "Could Not Create Working Directory")
                return 1

    gJobReport = JobReport(jobID, "JobWrapper")

    try:
        job = JobWrapper(jobID, gJobReport)
        job.initialize(arguments)  # initialize doesn't return S_OK/S_ERROR
    except Exception as exc:  # pylint: disable=broad-except
        gLogger.exception("JobWrapper failed the initialization phase", lException=exc)
        rescheduleResult = rescheduleFailedJob(
            jobID=jobID, minorStatus=JobMinorStatus.JOB_WRAPPER_INITIALIZATION, jobReport=gJobReport
        )
        job.sendJobAccounting(status=rescheduleResult, minorStatus=JobMinorStatus.JOB_WRAPPER_INITIALIZATION)
        return 1

    if "InputSandbox" in arguments["Job"]:
        gJobReport.commit()
        try:
            result = job.transferInputSandbox(arguments["Job"]["InputSandbox"])
            if not result["OK"]:
                gLogger.warn(result["Message"])
                raise JobWrapperError(result["Message"])
        except JobWrapperError:
            gLogger.exception("JobWrapper failed to download input sandbox")
            rescheduleResult = rescheduleFailedJob(
                jobID=jobID, minorStatus=JobMinorStatus.DOWNLOADING_INPUT_SANDBOX, jobReport=gJobReport
            )
            job.sendJobAccounting(status=rescheduleResult, minorStatus=JobMinorStatus.DOWNLOADING_INPUT_SANDBOX)
            return 1
        except Exception as exc:  # pylint: disable=broad-except
            gLogger.exception("JobWrapper raised exception while downloading input sandbox", lException=exc)
            rescheduleResult = rescheduleFailedJob(
                jobID=jobID, minorStatus=JobMinorStatus.DOWNLOADING_INPUT_SANDBOX, jobReport=gJobReport
            )
            job.sendJobAccounting(status=rescheduleResult, minorStatus=JobMinorStatus.DOWNLOADING_INPUT_SANDBOX)
            return 1
    else:
        gLogger.verbose("Job has no InputSandbox requirement")

    gJobReport.commit()

    if "InputData" in arguments["Job"]:
        if arguments["Job"]["InputData"]:
            try:
                result = job.resolveInputData()
                if not result["OK"]:
                    gLogger.warn(result["Message"])
                    raise JobWrapperError(result["Message"])
            except JobWrapperError:
                gLogger.exception("JobWrapper failed to resolve input data")
                rescheduleResult = rescheduleFailedJob(
                    jobID=jobID, minorStatus=JobMinorStatus.INPUT_DATA_RESOLUTION, jobReport=gJobReport
                )
                job.sendJobAccounting(status=rescheduleResult, minorStatus=JobMinorStatus.INPUT_DATA_RESOLUTION)
                return 1
            except Exception as exc:  # pylint: disable=broad-except
                gLogger.exception("JobWrapper raised exception while resolving input data", lException=exc)
                rescheduleResult = rescheduleFailedJob(
                    jobID=jobID, minorStatus=JobMinorStatus.INPUT_DATA_RESOLUTION, jobReport=gJobReport
                )
                job.sendJobAccounting(status=rescheduleResult, minorStatus=JobMinorStatus.INPUT_DATA_RESOLUTION)
                return 1
        else:
            gLogger.verbose("Job has a null InputData requirement:")
            gLogger.verbose(arguments)
    else:
        gLogger.verbose("Job has no InputData requirement")

    gJobReport.commit()

    try:
        result = job.execute()
        if not result["OK"]:
            gLogger.error("Failed to execute job", result["Message"])
            raise JobWrapperError((result["Message"], result["Errno"]))
    except JobWrapperError as exc:
        if exc.value[1] == 0 or str(exc.value[0]) == "0":
            gLogger.verbose("JobWrapper exited with status=0 after execution")
        if exc.value[1] == DErrno.EWMSRESC:
            gLogger.warn("Asked to reschedule job")
            rescheduleResult = rescheduleFailedJob(
                jobID=jobID, minorStatus=JobMinorStatus.JOB_WRAPPER_EXECUTION, jobReport=gJobReport
            )
            job.sendJobAccounting(status=rescheduleResult, minorStatus=JobMinorStatus.JOB_WRAPPER_EXECUTION)
            return 1
        gLogger.exception("Job failed in execution phase")
        gJobReport.setJobParameter("Error Message", repr(exc), sendFlag=False)
        gJobReport.setJobStatus(
            status=JobStatus.FAILED, minorStatus=JobMinorStatus.EXCEPTION_DURING_EXEC, sendFlag=False
        )
        job.sendFailoverRequest()
        job.sendJobAccounting(status=JobStatus.FAILED, minorStatus=JobMinorStatus.EXCEPTION_DURING_EXEC)
        return 1
    except Exception as exc:  # pylint: disable=broad-except
        gLogger.exception("Job raised exception during execution phase", lException=exc)
        gJobReport.setJobParameter("Error Message", repr(exc), sendFlag=False)
        gJobReport.setJobStatus(
            status=JobStatus.FAILED, minorStatus=JobMinorStatus.EXCEPTION_DURING_EXEC, sendFlag=False
        )
        job.sendFailoverRequest()
        job.sendJobAccounting(status=JobStatus.FAILED, minorStatus=JobMinorStatus.EXCEPTION_DURING_EXEC)
        return 1

    if "OutputSandbox" in arguments["Job"] or "OutputData" in arguments["Job"]:
        try:
            result = job.processJobOutputs()
            if not result["OK"]:
                gLogger.warn(result["Message"])
                raise JobWrapperError(result["Message"])
        except JobWrapperError as exc:
            gLogger.exception("JobWrapper failed to process output files")
            gJobReport.setJobParameter("Error Message", repr(exc), sendFlag=False)
            gJobReport.setJobStatus(
                status=JobStatus.FAILED, minorStatus=JobMinorStatus.UPLOADING_JOB_OUTPUTS, sendFlag=False
            )
            job.sendFailoverRequest()
            job.sendJobAccounting(status=JobStatus.FAILED, minorStatus=JobMinorStatus.UPLOADING_JOB_OUTPUTS)

            return 2
        except Exception as exc:  # pylint: disable=broad-except
            gLogger.exception("JobWrapper raised exception while processing output files", lException=exc)
            gJobReport.setJobParameter("Error Message", repr(exc), sendFlag=False)
            gJobReport.setJobStatus(
                status=JobStatus.FAILED, minorStatus=JobMinorStatus.UPLOADING_JOB_OUTPUTS, sendFlag=False
            )
            job.sendFailoverRequest()
            job.sendJobAccounting(status=JobStatus.FAILED, minorStatus=JobMinorStatus.UPLOADING_JOB_OUTPUTS)
            return 2
    else:
        gLogger.verbose("Job has no OutputData or OutputSandbox requirement")

    try:
        # Failed jobs will return !=0 / successful jobs will return 0
        return job.finalize()
    except Exception as exc:  # pylint: disable=broad-except
        gLogger.exception("JobWrapper raised exception during the finalization phase", lException=exc)
        return 2


##########################################################


ret = -3
try:
    jsonFileName = os.path.realpath(__file__) + ".json"
    with open(jsonFileName, "r") as f:
        jobArgsFromJSON = json.loads(f.readlines()[0])
    jobArgs = ast.literal_eval(jobArgsFromJSON)
    if not isinstance(jobArgs, dict):
        raise TypeError("jobArgs is of type %s" % type(jobArgs))
    if "Job" not in jobArgs:
        raise ValueError("jobArgs does not contain 'Job' key: %s" % str(jobArgs))
    ret = execute(jobArgs)
    gJobReport.commit()
except Exception as exc:  # pylint: disable=broad-except
    gLogger.exception("JobWrapperTemplate exception", lException=exc)
    try:
        gJobReport.commit()
        ret = -1
    except Exception as exc:  # pylint: disable=broad-except
        gLogger.exception("Could not commit the job report", lException=exc)
        ret = -2

sys.exit(ret)



"""Ensure the files have been migrated to tape."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
from pprint import pformat

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.Resources.Storage.StorageElement import StorageElement

__RCSID__ = "$Id$"


class CheckMigration(OperationHandlerBase):
    """CheckMigration operation handler."""

    def __init__(self, operation=None, csPath=None):
        """Initialize the ArchifeFiles handler.

        :param self: self reference
        :param Operation operation: Operation instance
        :param string csPath: CS path for this handler
        """
        OperationHandlerBase.__init__(self, operation, csPath)
        self.waitingFiles = []

    def __call__(self):
        """Process the CheckMigration operation."""
        try:
            self._run()
        except Exception as e:
            self.log.exception("Failed to execute CheckMigration", repr(e), lException=e)
            return S_ERROR(str(e))
        return S_OK()

    def _run(self):
        """Check for migration bit, set file done when migrated."""
        self.waitingFiles = self.getWaitingFilesList()
        self.log.notice("Waiting files:", len(self.waitingFiles))
        targetSESet = set(self.operation.targetSEList)
        self.log.notice("Target SEs:", ",".join(targetSESet))
        migrated = True
        for opFile in self.waitingFiles:
            self.log.notice("Checking:", opFile.LFN)
            for targetSE in targetSESet:
                se = StorageElement(targetSE)
                if not se.status()["TapeSE"]:
                    migrated = True and migrated
                    continue
                metaData = returnSingleResult(se.getFileMetadata(opFile.LFN))
                self.log.debug("MetaData: %s" % pformat(metaData))
                if not metaData["OK"]:
                    self.log.error("Failed to get metadata:", "%s: %s" % (opFile.LFN, metaData["Message"]))
                    migrated = False
                    continue
                migrated = metaData["Value"].get("Migrated", 0) == 1 and migrated
            if migrated:
                self.log.notice("File has been migrated:", opFile.LFN)
                opFile.Status = "Done"
            else:
                self.log.notice("File has NOT been migrated:", opFile.LFN)
                now = datetime.datetime.utcnow().replace(microsecond=0)
                extraDelay = datetime.timedelta(minutes=20)
                self.request.NotBefore = now + extraDelay



"""
ElasticSearch wrapper
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import logging
from cmreslogging.handlers import CMRESHandler

from DIRAC.Resources.LogBackends.AbstractBackend import AbstractBackend


class ElasticSearchBackend(AbstractBackend):
    """
    ElasticsearchBackend is used to create an abstraction of the handler and the formatter concepts from logging.
    Here, we have a CMRESHandler which is part of an external library named 'cmreslogging' based on 'logging'.
    CMRESHandler is a specific handler created to send log records to an ElasticSearch DB. It does not need a Formatter
    object.
    """

    def __init__(self, backendParams=None):
        """
        CMRESHandler needs, at least, a hostname, a username, a password, a port and a specific index
        from the ElasticSearch DB to send log records.
        """
        # We give a format containing only asctime to add the field in elasticsearch
        # asctime is not created at the initialization of the LogRecords but built in the format process
        if not backendParams:
            backendParams = {}
        backendParams["Format"] = "%(asctime)s"

        super(ElasticSearchBackend, self).__init__(CMRESHandler, logging.Formatter, backendParams)

    def _setHandlerParameters(self, backendParams=None):
        """
        Get the handler parameters from the backendParams.
        The keys of handlerParams should correspond to the parameter names of the associated handler.
        The method should be overridden in every backend that needs handler parameters.
        The method should be called before creating the handler object.

        :param dict parameters: parameters of the backend. ex: {'FileName': file.log}
        """
        # fixed parameters
        self._handlerParams["use_ssl"] = True
        self._handlerParams["verify_ssl"] = True
        self._handlerParams["auth_type"] = CMRESHandler.AuthType.NO_AUTH

        # variable parameters
        self._handlerParams["es_index_name"] = ""
        self._handlerParams["buffer_size"] = 1000
        self._handlerParams["flush_frequency_in_sec"] = 1
        user = None
        password = None
        host = ""
        port = 9203

        if backendParams is not None:
            self._handlerParams["es_index_name"] = backendParams.get("Index", self._handlerParams["es_index_name"])
            self._handlerParams["buffer_size"] = backendParams.get("BufferSize", self._handlerParams["buffer_size"])
            self._handlerParams["flush_frequency_in_sec"] = backendParams.get(
                "FlushTime", self._handlerParams["flush_frequency_in_sec"]
            )

            user = backendParams.get("User", user)
            password = backendParams.get("Password", password)
            if user is not None and password is not None:
                self._handlerParams["auth_type"] = CMRESHandler.AuthType.BASIC_AUTH
                self._handlerParams["auth_details"] = (user, password)

            host = backendParams.get("Host", host)
            port = int(backendParams.get("Port", port))

        self._handlerParams["hosts"] = [{"host": host, "port": port}]



""" Client for the SandboxStore.
    Will connect to the WorkloadManagement/SandboxStore service.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import tarfile
import hashlib
import tempfile
import re
from six import BytesIO, StringIO

from DIRAC import gLogger, S_OK, S_ERROR, gConfig

from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Utilities.File import getGlobbedTotalSize
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup


class SandboxStoreClient(object):

    __validSandboxTypes = ("Input", "Output")
    __smdb = None

    def __init__(self, rpcClient=None, transferClient=None, smdb=False, **kwargs):
        """Constructor

        :param object rpcClient: SandboxStore service client (None by default)
        :param object transferClient: client to upload/download sandboxes (None by default)
        :param object smdb: SandboxMetadataDB object, or
                            True if SandboxMetadataDB is to be instantiated for direct access or
                            False if no direct access to the SandboxMetadataDB is done (default)

        """

        self.__serviceName = "WorkloadManagement/SandboxStore"
        self.__rpcClient = rpcClient
        self.__transferClient = transferClient
        self.__kwargs = kwargs
        self.__vo = None
        SandboxStoreClient.__smdb = smdb
        if "delegatedGroup" in kwargs:
            self.__vo = getVOForGroup(kwargs["delegatedGroup"])
        if SandboxStoreClient.__smdb is True:
            try:
                from DIRAC.WorkloadManagementSystem.DB.SandboxMetadataDB import SandboxMetadataDB

                SandboxStoreClient.__smdb = SandboxMetadataDB()
                result = SandboxStoreClient.__smdb._getConnection()  # pylint: disable=protected-access
                if not result["OK"]:
                    SandboxStoreClient.__smdb = False
                else:
                    result["Value"].close()
            except (ImportError, RuntimeError, AttributeError):
                SandboxStoreClient.__smdb = False

    def __getRPCClient(self):
        """Get an RPC client for SB service"""
        if self.__rpcClient:
            return self.__rpcClient
        else:
            return RPCClient(self.__serviceName, **self.__kwargs)

    def __getTransferClient(self):
        """Get RPC client for TransferClient"""
        if self.__transferClient:
            return self.__transferClient
        else:
            return TransferClient(self.__serviceName, **self.__kwargs)

    # Upload sandbox to jobs and pilots

    def uploadFilesAsSandboxForJob(self, fileList, jobId, sbType, sizeLimit=0):
        """Upload SB for a job"""
        if sbType not in self.__validSandboxTypes:
            return S_ERROR("Invalid Sandbox type %s" % sbType)
        return self.uploadFilesAsSandbox(fileList, sizeLimit, assignTo={"Job:%s" % jobId: sbType})

    def uploadFilesAsSandboxForPilot(self, fileList, jobId, sbType, sizeLimit=0):
        """Upload SB for a pilot"""
        if sbType not in self.__validSandboxTypes:
            return S_ERROR("Invalid Sandbox type %s" % sbType)
        return self.uploadFilesAsSandbox(fileList, sizeLimit, assignTo={"Pilot:%s" % jobId: sbType})

    # Upload generic sandbox

    def uploadFilesAsSandbox(self, fileList, sizeLimit=0, assignTo=None):
        """Send files in the fileList to a Sandbox service for the given jobID.
        This is the preferable method to upload sandboxes.

        a fileList item can be:
          - a string, which is an lfn name
          - a file name (real), that is supposed to be on disk, in the current directory
          - a fileObject that should be a BytesIO type of object

        Parameters:
          - assignTo : Dict containing { 'Job:<jobid>' : '<sbType>', ... }
        """
        errorFiles = []
        files2Upload = []
        if assignTo is None:
            assignTo = {}

        for key in assignTo:
            if assignTo[key] not in self.__validSandboxTypes:
                return S_ERROR("Invalid sandbox type %s" % assignTo[key])

        if not isinstance(fileList, (list, tuple)):
            return S_ERROR("fileList must be a list or tuple!")

        for sFile in fileList:
            if isinstance(sFile, six.string_types):
                if re.search("^lfn:", sFile, flags=re.IGNORECASE):
                    pass
                else:
                    if os.path.exists(sFile):
                        files2Upload.append(sFile)
                    else:
                        errorFiles.append(sFile)

            elif isinstance(sFile, StringIO):
                files2Upload.append(sFile)
            else:
                return S_ERROR("Objects of type %s can't be part of InputSandbox" % type(sFile))

        if errorFiles:
            return S_ERROR("Failed to locate files: %s" % ", ".join(errorFiles))

        try:
            fd, tmpFilePath = tempfile.mkstemp(prefix="LDSB.")
            os.close(fd)
        except Exception as e:
            return S_ERROR("Cannot create temporary file: %s" % repr(e))

        with tarfile.open(name=tmpFilePath, mode="w|bz2") as tf:
            for sFile in files2Upload:
                if isinstance(sFile, six.string_types):
                    tf.add(os.path.realpath(sFile), os.path.basename(sFile), recursive=True)
                elif isinstance(sFile, StringIO):
                    tarInfo = tarfile.TarInfo(name="jobDescription.xml")
                    value = sFile.getvalue().encode()
                    tarInfo.size = len(value)
                    tf.addfile(tarinfo=tarInfo, fileobj=BytesIO(value))
                else:
                    return S_ERROR("Unknown type to upload: %s" % repr(sFile))

        if sizeLimit > 0:
            # Evaluate the compressed size of the sandbox
            if getGlobbedTotalSize(tmpFilePath) > sizeLimit:
                result = S_ERROR("Size over the limit")
                result["SandboxFileName"] = tmpFilePath
                return result

        oMD5 = hashlib.md5()
        with open(tmpFilePath, "rb") as fd:
            bData = fd.read(10240)
            while bData:
                oMD5.update(bData)
                bData = fd.read(10240)

        transferClient = self.__getTransferClient()
        result = transferClient.sendFile(tmpFilePath, ["%s.tar.bz2" % oMD5.hexdigest(), assignTo])
        result["SandboxFileName"] = tmpFilePath
        try:
            if result["OK"]:
                os.unlink(tmpFilePath)
        except OSError:
            pass
        return result

    ##############
    # Download sandbox

    def downloadSandbox(self, sbLocation, destinationDir="", inMemory=False, unpack=True):
        """
        Download a sandbox file and keep it in bundled form
        """
        if sbLocation.find("SB:") != 0:
            return S_ERROR("Invalid sandbox URL")
        sbLocation = sbLocation[3:]
        sbSplit = sbLocation.split("|")
        if len(sbSplit) < 2:
            return S_ERROR("Invalid sandbox URL")
        seName = sbSplit[0]
        sePFN = "|".join(sbSplit[1:])

        try:
            tmpSBDir = tempfile.mkdtemp(prefix="TMSB.")
        except IOError as e:
            return S_ERROR("Cannot create temporary file: %s" % repr(e))

        se = StorageElement(seName, vo=self.__vo)
        result = returnSingleResult(se.getFile(sePFN, localPath=tmpSBDir))

        if not result["OK"]:
            return result
        sbFileName = os.path.basename(sePFN)

        result = S_OK()
        tarFileName = os.path.join(tmpSBDir, sbFileName)

        if inMemory:
            try:
                with open(tarFileName, "rb") as tfile:
                    data = tfile.read()
            except IOError as e:
                return S_ERROR("Failed to read the sandbox archive: %s" % repr(e))
            finally:
                os.unlink(tarFileName)
                os.rmdir(tmpSBDir)
            return S_OK(data)

        # If destination dir is not specified use current working dir
        # If its defined ensure the dir structure is there
        if not destinationDir:
            destinationDir = os.getcwd()
        else:
            mkDir(destinationDir)

        if not unpack:
            result["Value"] = tarFileName
            return result

        try:
            sandboxSize = 0
            with tarfile.open(name=tarFileName, mode="r") as tf:
                for tarinfo in tf:
                    tf.extract(tarinfo, path=destinationDir)
                    sandboxSize += tarinfo.size
            # FIXME: here we return the size, but otherwise we always return the location: inconsistent
            # FIXME: looks like this size is used by the JobWrapper
            result["Value"] = sandboxSize
        except IOError as e:
            result = S_ERROR("Could not open bundle: %s" % repr(e))

        try:
            os.unlink(tarFileName)
            os.rmdir(tmpSBDir)
        except OSError as e:
            gLogger.warn("Could not remove temporary dir %s: %s" % (tmpSBDir, repr(e)))

        return result

    ##############
    # Jobs

    def getSandboxesForJob(self, jobId):
        """Download job sandbox"""
        return self.__getSandboxesForEntity("Job:%s" % jobId)

    def assignSandboxesToJob(self, jobId, sbList, ownerName="", ownerGroup="", eSetup=""):
        """Assign SB to a job"""
        return self.__assignSandboxesToEntity("Job:%s" % jobId, sbList, ownerName, ownerGroup, eSetup)

    def assignSandboxToJob(self, jobId, sbLocation, sbType, ownerName="", ownerGroup="", eSetup=""):
        """Assign SB to a job"""
        return self.__assignSandboxToEntity("Job:%s" % jobId, sbLocation, sbType, ownerName, ownerGroup, eSetup)

    def unassignJobs(self, jobIdList):
        """Unassign SB to a job"""
        if isinstance(jobIdList, six.integer_types):
            jobIdList = [jobIdList]
        entitiesList = []
        for jobId in jobIdList:
            entitiesList.append("Job:%s" % jobId)
        return self.__unassignEntities(entitiesList)

    def downloadSandboxForJob(self, jobId, sbType, destinationPath="", inMemory=False, unpack=True):
        """Download SB for a job"""
        result = self.__getSandboxesForEntity("Job:%s" % jobId)
        if not result["OK"]:
            return result
        sbDict = result["Value"]
        if sbType not in sbDict:
            return S_ERROR(
                "No %s sandbox found for job %s. " % (sbType, jobId)
                + "Possible causes are: the job does not exist, no sandbox was "
                "registered or you do not have permission to access it."
            )

        # If inMemory, ensure we return the newest sandbox only
        if inMemory:
            sbLocation = sbDict[sbType][-1]
            return self.downloadSandbox(sbLocation, destinationPath, inMemory, unpack)

        downloadedSandboxesLoc = []
        for sbLocation in sbDict[sbType]:
            result = self.downloadSandbox(sbLocation, destinationPath, inMemory, unpack)
            if not result["OK"]:
                return result
            downloadedSandboxesLoc.append(result["Value"])
        return S_OK(downloadedSandboxesLoc)

    ##############
    # Pilots

    def getSandboxesForPilot(self, pilotId):
        """Get SB for a pilot"""
        return self.__getSandboxesForEntity("Pilot:%s" % pilotId)

    def assignSandboxesToPilot(self, pilotId, sbList, ownerName="", ownerGroup="", eSetup=""):
        """Assign SB to a pilot"""
        return self.__assignSandboxesToEntity("Pilot:%s" % pilotId, sbList, ownerName, ownerGroup, eSetup)

    def assignSandboxToPilot(self, pilotId, sbLocation, sbType, ownerName="", ownerGroup="", eSetup=""):
        """Assign SB to a pilot"""
        return self.__assignSandboxToEntity("Pilot:%s" % pilotId, sbLocation, sbType, ownerName, ownerGroup, eSetup)

    def unassignPilots(self, pilotIdIdList):
        """Unassign SB to a pilot"""
        if isinstance(pilotIdIdList, six.integer_types):
            pilotIdIdList = [pilotIdIdList]
        entitiesList = []
        for pilotId in pilotIdIdList:
            entitiesList.append("Pilot:%s" % pilotId)
        return self.__unassignEntities(entitiesList)

    def downloadSandboxForPilot(self, jobId, sbType, destinationPath=""):
        """Download SB for a pilot"""
        result = self.__getSandboxesForEntity("Pilot:%s" % jobId)
        if not result["OK"]:
            return result
        sbDict = result["Value"]
        if sbType not in sbDict:
            return S_ERROR("No %s sandbox registered for pilot %s" % (sbType, jobId))

        downloadedSandboxesLoc = []
        for sbLocation in sbDict[sbType]:
            result = self.downloadSandbox(sbLocation, destinationPath)
            if not result["OK"]:
                return result
            downloadedSandboxesLoc.append(result["Value"])
        return S_OK(downloadedSandboxesLoc)

    ##############
    # Entities

    def __getSandboxesForEntity(self, eId):
        """
        Get the sandboxes assigned to jobs and the relation type
        """
        rpcClient = self.__getRPCClient()
        return rpcClient.getSandboxesAssignedToEntity(eId)

    def __assignSandboxesToEntity(self, eId, sbList, ownerName="", ownerGroup="", eSetup=""):
        """
        Assign sandboxes to a job.
        sbList must be a list of sandboxes and relation types
          sbList = [ ( "SB:SEName|SEPFN", "Input" ), ( "SB:SEName|SEPFN", "Output" ) ]
        """
        for sbT in sbList:
            if sbT[1] not in self.__validSandboxTypes:
                return S_ERROR("Invalid Sandbox type %s" % sbT[1])
        if SandboxStoreClient.__smdb and ownerName and ownerGroup:
            if not eSetup:
                eSetup = gConfig.getValue("/DIRAC/Setup", "Production")
            return SandboxStoreClient.__smdb.assignSandboxesToEntities({eId: sbList}, ownerName, ownerGroup, eSetup)
        rpcClient = self.__getRPCClient()
        return rpcClient.assignSandboxesToEntities({eId: sbList}, ownerName, ownerGroup, eSetup)

    def __assignSandboxToEntity(self, eId, sbLocation, sbType, ownerName="", ownerGroup="", eSetup=""):
        """
        Assign a sandbox to a job
          sbLocation is "SEName:SEPFN"
          sbType is Input or Output
        """
        return self.__assignSandboxesToEntity(eId, [(sbLocation, sbType)], ownerName, ownerGroup, eSetup)

    def __unassignEntities(self, eIdList):
        """
        Unassign a list of jobs of their respective sandboxes
        """
        rpcClient = self.__getRPCClient()
        return rpcClient.unassignEntities(eIdList)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-site-mask
# Author :  Stuart Paterson
########################################################################
"""
Get the list of sites enabled in the mask for job submission

Example:
  $ dirac-admin-get-site-mask
  LCG.CGG.fr
  LCG.CPPM.fr
  LCG.LAPP.fr
  LCG.LPSC.fr
  LCG.M3PEC.fr
  LCG.MSFG.fr
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit, gLogger
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()

    gLogger.setLevel("ALWAYS")

    result = diracAdmin.getSiteMask(printOutput=True, status="Active")
    if result["OK"]:
        DIRACExit(0)
    else:
        print(result["Message"])
        DIRACExit(2)


if __name__ == "__main__":
    main()



"""
Message Queue Handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import json
import logging
import socket

from DIRAC.Resources.MessageQueue.MQCommunication import createProducer


class MessageQueueHandler(logging.Handler):
    """
    MessageQueueHandler is a custom handler from logging.
    It has no equivalent in the standard logging library because it is linked to DIRAC.

    It is useful to send log messages to a destination, like the StreamHandler to a stream, the FileHandler to a file.
    Here, this handler send log messages to a message queue server.

    There is an assumption made that the formatter used is JsonFormatter
    """

    def __init__(self, queue):
        """
        Initialization of the MessageQueueHandler.

        :param queue: queue identifier in the configuration.
                      example: "mardirac3.in2p3.fr::Queues::TestQueue"
        """
        super(MessageQueueHandler, self).__init__()
        self.producer = None
        result = createProducer(queue)
        if result["OK"]:
            self.producer = result["Value"]
        else:
            # print because logging not available
            print("ERROR initializing MessageQueueHandler: %s" % result)
        self.hostname = socket.gethostname()

    def emit(self, record):
        """
        Add the record to the message queue.

        :param record: log record object
        """
        # add the hostname to the record
        record.hostname = self.hostname
        strRecord = self.format(record)
        if self.producer is not None:
            self.producer.put(json.loads(strRecord))



########################################################################
# File :   InProcessComputingElement.py
# Author : Stuart Paterson
########################################################################

""" The simplest of the "inner" CEs (meaning it's used by a jobAgent inside a pilot)

    A "InProcess" CE instance submits jobs in the current process.
    This is the standard "inner CE" invoked from the JobAgent, main alternative being the PoolCE
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import stat

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler

from DIRAC.Resources.Computing.ComputingElement import ComputingElement


class InProcessComputingElement(ComputingElement):

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(InProcessComputingElement, self).__init__(ceUniqueID)

        self.ceType = "InProcess"
        self.submittedJobs = 0
        self.runningJobs = 0

        self.processors = int(self.ceParameters.get("NumberOfProcessors", 1))
        self.ceParameters["MaxTotalJobs"] = 1

    #############################################################################
    def submitJob(self, executableFile, proxy=None, **kwargs):
        """Method to submit job (overriding base method).

        :param str executableFile: file to execute via systemCall.
                                   Normally the JobWrapperTemplate when invoked by the JobAgent.
        :param str proxy: the proxy used for running the job (the payload). It will be dumped to a file.
        """
        payloadEnv = dict(os.environ)
        payloadProxy = ""
        if proxy:
            self.log.verbose("Setting up proxy for payload")
            result = self.writeProxyToFile(proxy)
            if not result["OK"]:
                return result

            payloadProxy = result["Value"]  # payload proxy file location
            payloadEnv["X509_USER_PROXY"] = payloadProxy

            self.log.verbose("Starting process for monitoring payload proxy")
            result = gThreadScheduler.addPeriodicTask(
                self.proxyCheckPeriod, self._monitorProxy, taskArgs=(payloadProxy,), executions=0, elapsedTime=0
            )
            if result["OK"]:
                renewTask = result["Value"]
            else:
                self.log.warn("Failed to start proxy renewal task")
                renewTask = None

        self.submittedJobs += 1
        self.runningJobs += 1

        if not os.access(executableFile, 5):
            os.chmod(executableFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)
        cmd = os.path.abspath(executableFile)
        self.log.verbose("CE submission command:", cmd)
        result = systemCall(0, cmd, callbackFunction=self.sendOutput, env=payloadEnv)
        if payloadProxy:
            os.unlink(payloadProxy)

        if proxy and renewTask:
            gThreadScheduler.removeTask(renewTask)

        self.runningJobs -= 1

        ret = S_OK()

        if not result["OK"]:
            self.log.error("Fail to run InProcess", result["Message"])
        elif result["Value"][0] > 128:
            res = S_ERROR()
            # negative exit values are returned as 256 - exit
            res["Value"] = result["Value"][0] - 256  # yes, it's "correct"
            self.log.warn("InProcess Job Execution Failed")
            self.log.info("Exit status:", result["Value"])
            if res["Value"] == -2:
                error = "JobWrapper initialization error"
            elif res["Value"] == -1:
                error = "JobWrapper execution error"
            else:
                error = "InProcess Job Execution Failed"
            res["Message"] = error
            return res
        elif result["Value"][0] > 0:
            self.log.warn("Fail in payload execution")
            self.log.info("Exit status:", result["Value"][0])
            ret["PayloadFailed"] = result["Value"][0]
        else:
            self.log.debug("InProcess CE result OK")

        return ret

    #############################################################################
    def getCEStatus(self):
        """Method to return information on running and waiting jobs,
        as well as number of available processors
        """
        result = S_OK()

        result["SubmittedJobs"] = self.submittedJobs
        result["RunningJobs"] = self.runningJobs
        result["WaitingJobs"] = 0
        # processors
        result["AvailableProcessors"] = self.processors
        return result



""" Failover Transfer

    The failover transfer client exposes the following methods:
    - transferAndRegisterFile()
    - transferAndRegisterFileFailover()

    Initially these methods were developed inside workflow modules but
    have evolved to a generic 'transfer file with failover' client.

    The transferAndRegisterFile() method will correctly set registration
    requests in case of failure.

    The transferAndRegisterFileFailover() method will attempt to upload
    a file to a list of alternative SEs and set appropriate replication
    to the original target SE as well as the removal request for the
    temporary replica.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time

from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Utilities.DErrno import cmpError, EFCERR
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog


class FailoverTransfer(object):
    """.. class:: FailoverTransfer"""

    #############################################################################
    def __init__(self, requestObject=None, log=None, defaultChecksumType="ADLER32"):
        """Constructor function, can specify request object to instantiate
        FailoverTransfer or a new request object is created.
        """
        self.log = log
        if not self.log:
            self.log = gLogger.getSubLogger("FailoverTransfer")

        self.request = requestObject
        if not self.request:
            self.request = Request()
            self.request.RequestName = "noname_request"
            self.request.SourceComponent = "FailoverTransfer"

        self.defaultChecksumType = defaultChecksumType
        self.registrationProtocols = DMSHelpers().getRegistrationProtocols()

    #############################################################################
    def transferAndRegisterFile(
        self,
        fileName,
        localPath,
        lfn,
        destinationSEList,
        fileMetaDict,
        fileCatalog=None,
        masterCatalogOnly=False,
        retryUpload=False,
    ):
        """Performs the transfer and register operation with failover.

        :param filename: of absolute no use except for printing logs.
        :param localPath: path to the file locally
        :param lfn: LFN
        :param destinationSEList: list of possible destination for the file.
          Loop over it until one succeeds or we reach the end of it.
        :param fileMetaDict: file metadata for registration
        :param fileCatalog: list of catalogs to use (see :py:class:`DIRAC.DataManagementSystem.Client.DataManager`)
        :param masterCatalogOnly: use only master catalog (see :py:class:`DIRAC.DataManagementSystem.Client.DataManager`)
        :param retryUpload: if set to True, and there is only one output SE in destinationSEList, retry several times.

        """
        errorList = []
        fileGUID = fileMetaDict.get("GUID", None)
        fileChecksum = fileMetaDict.get("Checksum", None)

        for se in destinationSEList:

            # We put here some retry in case the problem comes from the FileCatalog
            # being unavailable. If it is, then the `hasAccess` call would fail,
            # and we would not make any failover request. So the only way is to wait a bit
            # This keeps the WN busy for a while, but at least we do not lose all the processing
            # time we just spent
            # This same retry path is taken if we only have one possible stage out SE
            # and retryUpload is True
            for sleeptime in (10, 60, 300, 600):
                self.log.info(
                    "Attempting dm.putAndRegister",
                    "('%s','%s','%s',guid='%s',catalog='%s', checksum = '%s')"
                    % (lfn, localPath, se, fileGUID, fileCatalog, fileChecksum),
                )

                result = DataManager(catalogs=fileCatalog, masterCatalogOnly=masterCatalogOnly).putAndRegister(
                    lfn, localPath, se, guid=fileGUID, checksum=fileChecksum
                )
                # retry on any failure
                if result["OK"]:
                    self.log.verbose(result)
                    break
                elif cmpError(result, EFCERR):
                    self.log.debug("transferAndRegisterFile: FC unavailable, retry")
                elif retryUpload and len(destinationSEList) == 1:
                    self.log.debug("transferAndRegisterFile: Failed uploading to the only SE, retry")
                else:
                    self.log.debug("dm.putAndRegister failed, but move to the next")
                    break
                time.sleep(sleeptime)

            if not result["OK"]:
                self.log.error("dm.putAndRegister failed with message", result["Message"])
                errorList.append(result["Message"])
                continue

            if not result["Value"]["Failed"]:
                self.log.info("dm.putAndRegister successfully uploaded and registered", "%s to %s" % (fileName, se))
                return S_OK({"uploadedSE": se, "lfn": lfn})

            # Now we know something went wrong
            self.log.warn("Didn't manage to do everything, now adding requests for the missing operation")

            errorDict = result["Value"]["Failed"][lfn]
            if "register" not in errorDict:
                self.log.error("dm.putAndRegister failed with unknown error", str(errorDict))
                errorList.append("Unknown error while attempting upload to %s" % se)
                continue

            # fileDict = errorDict['register']
            # Therefore the registration failed but the upload was successful
            if not fileCatalog:
                fileCatalog = ""

            if masterCatalogOnly:
                fileCatalog = FileCatalog().getMasterCatalogNames()["Value"]

            result = self._setRegistrationRequest(lfn, se, fileMetaDict, fileCatalog)
            if not result["OK"]:
                self.log.error("Failed to set registration request", "SE %s and metadata: \n%s" % (se, fileMetaDict))
                errorList.append(
                    "Failed to set registration request for: SE %s and metadata: \n%s" % (se, fileMetaDict)
                )
                continue
            else:
                self.log.info(
                    "Successfully set registration request", "for: SE %s and metadata: \n%s" % (se, fileMetaDict)
                )
                metadata = {}
                metadata["filedict"] = fileMetaDict
                metadata["uploadedSE"] = se
                metadata["lfn"] = lfn
                metadata["registration"] = "request"
                return S_OK(metadata)

        self.log.error("Failed to upload output data file", "Encountered %s errors" % len(errorList))
        return S_ERROR("Failed to upload output data file")

    #############################################################################
    def transferAndRegisterFileFailover(
        self,
        fileName,
        localPath,
        lfn,
        targetSE,
        failoverSEList,
        fileMetaDict,
        fileCatalog=None,
        masterCatalogOnly=False,
    ):
        """Performs the transfer and register operation to failover storage and sets the
        necessary replication and removal requests to recover.
        """
        failover = self.transferAndRegisterFile(
            fileName, localPath, lfn, failoverSEList, fileMetaDict, fileCatalog, masterCatalogOnly=masterCatalogOnly
        )
        if not failover["OK"]:
            self.log.error("Could not upload file to failover SEs", failover["Message"])
            return failover

        # set removal requests and replication requests
        result = self._setFileReplicationRequest(lfn, targetSE, fileMetaDict, sourceSE=failover["Value"]["uploadedSE"])
        if not result["OK"]:
            self.log.error("Could not set file replication request", result["Message"])
            return result

        lfn = failover["Value"]["lfn"]
        failoverSE = failover["Value"]["uploadedSE"]
        self.log.info("Attempting to set replica removal request", "for LFN %s at failover SE %s" % (lfn, failoverSE))
        result = self._setReplicaRemovalRequest(lfn, failoverSE)
        if not result["OK"]:
            self.log.error("Could not set removal request", result["Message"])
            return result

        return S_OK({"uploadedSE": failoverSE, "lfn": lfn})

    def getRequest(self):
        """get the accumulated request object"""
        return self.request

    def commitRequest(self):
        """Send request to the Request Management Service"""
        if self.request.isEmpty():
            return S_OK()

        isValid = RequestValidator().validate(self.request)
        if not isValid["OK"]:
            return S_ERROR("Failover request is not valid: %s" % isValid["Message"])
        else:
            requestClient = ReqClient()
            result = requestClient.putRequest(self.request)
            return result

    #############################################################################
    def _setFileReplicationRequest(self, lfn, targetSE, fileMetaDict, sourceSE=""):
        """Sets a registration request."""
        self.log.info("Setting ReplicateAndRegister request", "for %s to %s" % (lfn, targetSE))

        transfer = Operation()
        transfer.Type = "ReplicateAndRegister"
        transfer.TargetSE = targetSE
        if sourceSE:
            transfer.SourceSE = sourceSE

        trFile = File()
        trFile.LFN = lfn

        cksm = fileMetaDict.get("Checksum", None)
        cksmType = fileMetaDict.get("ChecksumType", self.defaultChecksumType)
        if cksm and cksmType:
            trFile.Checksum = cksm
            trFile.ChecksumType = cksmType
        size = fileMetaDict.get("Size", 0)
        if size:
            trFile.Size = size
        guid = fileMetaDict.get("GUID", "")
        if guid:
            trFile.GUID = guid

        transfer.addFile(trFile)

        self.request.addOperation(transfer)

        return S_OK()

    #############################################################################
    def _setRegistrationRequest(self, lfn, targetSE, fileDict, catalog):
        """Sets a registration request

        :param str lfn: LFN
        :param list se: list of SE (or just string)
        :param list catalog: list (or string) of catalogs to use
        :param dict fileDict: file metadata
        """
        self.log.info("Setting registration request", "for %s at %s." % (lfn, targetSE))

        if not isinstance(catalog, list):
            catalog = [catalog]

        for cat in catalog:

            register = Operation()
            register.Type = "RegisterFile"
            register.Catalog = cat
            register.TargetSE = targetSE

            regFile = File()
            regFile.LFN = lfn
            regFile.Checksum = fileDict.get("Checksum", "")
            regFile.ChecksumType = fileDict.get("ChecksumType", self.defaultChecksumType)
            regFile.Size = fileDict.get("Size", 0)
            regFile.GUID = fileDict.get("GUID", "")

            se = StorageElement(targetSE)
            res = returnSingleResult(se.getURL(lfn, self.registrationProtocols))
            if not res["OK"]:
                self.log.error("Unable to get PFN for LFN", res["Message"])
                return res
            regFile.PFN = res["Value"]

            register.addFile(regFile)
            self.request.addOperation(register)

        return S_OK()

    #############################################################################
    def _setReplicaRemovalRequest(self, lfn, se):
        """Sets a removal request for a replica.

        :param str lfn: LFN
        :param se:
        """
        if isinstance(se, str):
            se = ",".join([se.strip() for se in se.split(",") if se.strip()])

        removeReplica = Operation()

        removeReplica.Type = "RemoveReplica"
        removeReplica.TargetSE = se

        replicaToRemove = File()
        replicaToRemove.LFN = lfn

        removeReplica.addFile(replicaToRemove)

        self.request.addOperation(removeReplica)
        return S_OK()

    #############################################################################
    def _setFileRemovalRequest(self, lfn, se="", pfn=""):
        """Sets a removal request for a file including all replicas."""
        remove = Operation()
        remove.Type = "RemoveFile"
        if se:
            remove.TargetSE = se
        rmFile = File()
        rmFile.LFN = lfn
        if pfn:
            rmFile.PFN = pfn
        remove.addFile(rmFile)
        self.request.addOperation(remove)
        return S_OK()



#!/usr/bin/env python
"""
Do the initial installation and configuration of a DIRAC service based on tornado
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import exit as DIRACexit
from DIRAC import gConfig, gLogger, S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities.Extensions import extensionsByPriority
from DIRAC.FrameworkSystem.Utilities import MonitoringUtilities


overwrite = False


def setOverwrite(opVal):
    global overwrite
    overwrite = True
    return S_OK()


module = ""
specialOptions = {}


def setModule(optVal):
    global specialOptions, module
    specialOptions["Module"] = optVal
    module = optVal
    return S_OK()


def setSpecialOption(optVal):
    global specialOptions
    option, value = optVal.split("=")
    specialOptions[option] = value
    return S_OK()


@Script()
def main():
    global overwrite
    global specialOptions
    global module
    global specialOptions

    from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller

    gComponentInstaller.exitOnError = True

    Script.registerSwitch("w", "overwrite", "Overwrite the configuration in the global CS", setOverwrite)
    Script.registerSwitch("m:", "module=", "Python module name for the component code", setModule)
    Script.registerSwitch("p:", "parameter=", "Special component option ", setSpecialOption)
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        (
            "System/Component: Full component name (ie: WorkloadManagement/Matcher)",
            "System:           Name of the DIRAC system (ie: WorkloadManagement)",
        )
    )
    Script.registerArgument(" Component:        Name of the DIRAC service (ie: Matcher)", mandatory=False)
    Script.parseCommandLine()
    args = Script.getPositionalArgs()

    if len(args) == 1:
        args = args[0].split("/")

    if len(args) != 2:
        Script.showHelp()
        DIRACexit(1)

    system = args[0]
    component = args[1]

    result = gComponentInstaller.addDefaultOptionsToCS(
        gConfig,
        "service",
        system,
        component,
        extensionsByPriority(),
        specialOptions=specialOptions,
        overwrite=overwrite,
    )

    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)

    result = gComponentInstaller.addTornadoOptionsToCS(gConfig)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)

    result = gComponentInstaller.installTornado()
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)

    gLogger.notice("Successfully installed component %s in %s system, now setting it up" % (component, system))
    result = gComponentInstaller.setupTornadoService(system, component, extensionsByPriority(), module)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)

    result = MonitoringUtilities.monitorInstallation("service", system, component, module)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    gLogger.notice("Successfully completed the installation of %s/%s" % (system, component))
    DIRACexit()


if __name__ == "__main__":
    main()



"""
DIRAC Wrapper to execute python and system commands with a wrapper, that might
set a timeout.
3 FUNCTIONS are provided:

     - shellCall( iTimeOut, cmdSeq, callbackFunction = None, env = None ):
       it uses subprocess.Popen class with "shell = True".
       If cmdSeq is a string, it specifies the command string to execute through
       the shell.  If cmdSeq is a sequence, the first item specifies the command
       string, and any additional items will be treated as additional shell arguments.

     - systemCall( iTimeOut, cmdSeq, callbackFunction = None, env = None ):
       it uses subprocess.Popen class with "shell = False".
       cmdSeq should be a string, or a sequence of program arguments.

       stderr and stdout are piped. callbackFunction( pipeId, line ) can be
       defined to process the stdout (pipeId = 0) and stderr (pipeId = 1) as
       they are produced

       They return a DIRAC.ReturnValue dictionary with a tuple in Value
       ( returncode, stdout, stderr ) the tuple will also be available upon
       timeout error or buffer overflow error.

     - pythonCall( iTimeOut, function, \\*stArgs, \\*\\*stKeyArgs )
       calls function with given arguments within a timeout Wrapper
       should be used to wrap third party python functions

"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from multiprocessing import Process, Manager
import threading
import time
import os
import sys
import subprocess32 as subprocess
import signal
import psutil

try:
    import selectors
except ImportError:
    import selectors2 as selectors

# Very Important:
#  Here we can not import directly from DIRAC, since this file it is imported
#  at initialization time therefore the full path is necessary
# from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR

# from DIRAC import gLogger
from DIRAC.FrameworkSystem.Client.Logger import gLogger

__RCSID__ = "$Id$"

USE_WATCHDOG = False


class Watchdog(object):
    """
    .. class Watchdog

    timeout watchdog decorator
    """

    def __init__(self, func, args=None, kwargs=None):
        """c'tor"""
        self.func = func if callable(func) else None
        self.args = args if args else tuple()
        self.kwargs = kwargs if kwargs else {}
        self.start = self.end = self.pid = None
        self.rwEvent = threading.Event()
        self.rwEvent.clear()
        self.__watchdogThread = None
        self.manager = Manager()
        self.s_ok_error = self.manager.dict()
        self.__executor = Process(target=self.run_func, args=(self.s_ok_error,))

    def run_func(self, s_ok_error):
        """subprocess target

        :param Pipe pipe: pipe used for communication
        """
        try:
            ret = self.func(*self.args, **self.kwargs)
            # set rw event
            self.rwEvent.set()
            for k in ret:
                s_ok_error[k] = ret[k]
        except Exception as error:
            s_ok_error["OK"] = False
            s_ok_error["Message"] = str(error)
        finally:
            # clear rw event
            self.rwEvent.clear()

    def watchdog(self):
        """watchdog thread target"""
        while True:
            if self.rwEvent.is_set() or time.time() < self.end:
                time.sleep(5)
            else:
                break
        if not self.__executor.is_alive():
            return
        else:
            # wait until r/w operation finishes
            while self.rwEvent.is_set():
                time.sleep(5)
                continue
            # SIGTERM
            os.kill(self.pid, signal.SIGTERM)
            time.sleep(5)
            # SIGKILL
            if self.__executor.is_alive():
                os.kill(self.pid, signal.SIGKILL)

    def __call__(self, timeout=0):
        """decorator execution"""
        timeout = int(timeout)
        ret = {"OK": True, "Value": ""}
        if timeout:
            self.start = int(time.time())
            self.end = self.start + timeout + 2
            self.__watchdogThread = threading.Thread(target=self.watchdog)
            self.__watchdogThread.daemon = True
            self.__watchdogThread.start()
            ret = {"OK": False, "Message": "Timeout after %s seconds" % timeout, "Value": (1, "", "")}
        try:
            self.__executor.start()
            time.sleep(0.5)
            self.pid = self.__executor.pid
            if timeout:
                self.__executor.join(timeout)
            else:
                self.__executor.join()
            # get results if any, block watchdog by setting rwEvent
            if not self.__executor.is_alive():
                self.rwEvent.set()
                for k in self.s_ok_error.keys():
                    ret[k] = self.s_ok_error[k]
                self.rwEvent.clear()
        except Exception as error:
            return {"OK": False, "Message": str(error), "Value": (2, "", "")}
        return ret


class Subprocess(object):
    """
    .. class:: Subprocess

    """

    def __init__(self, timeout=False, bufferLimit=52428800):
        """c'tor

        :param int timeout: timeout in seconds
        :param int bufferLimit: buffer size, default 5MB
        """
        self.log = gLogger.getSubLogger("Subprocess")
        self.timeout = False
        try:
            self.changeTimeout(timeout)
            self.bufferLimit = int(bufferLimit)  # 5MB limit for data
        except Exception as x:
            self.log.exception("Failed initialisation of Subprocess object")
            raise x

        self.child = None
        self.childPID = 0
        self.childKilled = False
        self.callback = None
        self.bufferList = []
        self.cmdSeq = []

    def changeTimeout(self, timeout):
        """set the time out limit to :timeout: seconds

        :param int timeout: time out in seconds
        """
        self.timeout = int(timeout)
        if self.timeout == 0:
            self.timeout = False
        # self.log.debug( 'Timeout set to', timeout )

    def __readFromFD(self, fd, baseLength=0):
        """read from file descriptior :fd:

        :param fd: file descriptior
        :param int baseLength: ???
        """
        dataString = ""
        redBuf = " "
        while len(redBuf) > 0:
            redBuf = os.read(fd, 8192).decode()
            dataString += redBuf
            if len(dataString) + baseLength > self.bufferLimit:
                self.log.error(
                    "Maximum output buffer length reached",
                    "First and last data in buffer: \n%s \n....\n %s " % (dataString[:100], dataString[-100:]),
                )
                retDict = S_ERROR(
                    "Reached maximum allowed length (%d bytes) " "for called function return value" % self.bufferLimit
                )
                retDict["Value"] = dataString
                return retDict

        return S_OK(dataString)

    def __executePythonFunction(self, function, writePipe, *stArgs, **stKeyArgs):
        """
        execute function :funtion: using :stArgs: and :stKeyArgs:

        """

        from DIRAC.Core.Utilities import DEncode

        try:
            os.write(writePipe, DEncode.encode(S_OK(function(*stArgs, **stKeyArgs))))
        except OSError as x:
            if str(x) == "[Errno 32] Broken pipe":
                # the parent has died
                pass
        except Exception as x:
            self.log.exception("Exception while executing", function.__name__)
            os.write(writePipe, DEncode.encode(S_ERROR(str(x))))
            # HACK: Allow some time to flush logs
            time.sleep(1)
        try:
            os.close(writePipe)
        finally:
            os._exit(0)

    def __selectFD(self, readSeq, timeout=False):
        """select file descriptor from :readSeq: list"""
        validList = []
        for fd in readSeq:
            try:
                os.fstat(fd)
                validList.append(fd)
            except OSError:
                pass
        if not validList:
            return False
        sel = selectors.DefaultSelector()
        for socket in validList:
            sel.register(socket, selectors.EVENT_READ)
        events = sel.select(timeout=timeout or self.timeout or None)
        return [key.fileobj for key, event in events if event & selectors.EVENT_READ]

    def __killPid(self, pid, sig=9):
        """send signal :sig: to process :pid:

        :param int pid: process id
        :param int sig: signal to send, default 9 (SIGKILL)
        """
        try:
            os.kill(pid, sig)
        except Exception as x:
            if str(x) != "[Errno 3] No such process":
                self.log.exception("Exception while killing timed out process")
                raise x

    def __poll(self, pid):
        """wait for :pid:"""
        try:
            return os.waitpid(pid, os.WNOHANG)
        except os.error:
            if self.childKilled:
                return False
            return None

    def killChild(self, recursive=True):
        """kill child process

        FIXME: this can easily be rewritten just using this recipe:
        https://psutil.readthedocs.io/en/latest/index.html#kill-process-tree

        :param boolean recursive: flag to kill all descendants
        """
        if self.childPID < 1:
            self.log.error("Could not kill child", "Child PID is %s" % self.childPID)
            return -1
        os.kill(self.childPID, signal.SIGSTOP)
        if recursive:
            for gcpid in getChildrenPIDs(self.childPID, lambda cpid: os.kill(cpid, signal.SIGSTOP)):
                try:
                    os.kill(gcpid, signal.SIGKILL)
                    self.__poll(gcpid)
                except Exception:
                    pass
        self.__killPid(self.childPID)

        # HACK to avoid python bug
        # self.child.wait()
        exitStatus = self.__poll(self.childPID)
        i = 0
        while exitStatus is None and i < 1000:
            i += 1
            time.sleep(0.000001)
            exitStatus = self.__poll(self.childPID)
        try:
            exitStatus = os.waitpid(self.childPID, 0)
        except os.error:
            pass
        self.childKilled = True
        if exitStatus is None:
            return exitStatus
        return exitStatus[1]

    def pythonCall(self, function, *stArgs, **stKeyArgs):
        """call python function :function: with :stArgs: and :stKeyArgs:"""

        from DIRAC.Core.Utilities import DEncode

        self.log.verbose("pythonCall:", function.__name__)

        readFD, writeFD = os.pipe()
        pid = os.fork()
        self.childPID = pid
        if pid == 0:
            os.close(readFD)
            self.__executePythonFunction(function, writeFD, *stArgs, **stKeyArgs)
            # FIXME: the close it is done at __executePythonFunction, do we need it here?
            os.close(writeFD)
        else:
            os.close(writeFD)
            readSeq = self.__selectFD([readFD])
            if not readSeq:
                return S_ERROR("Can't read from call %s" % (function.__name__))
            try:
                if len(readSeq) == 0:
                    self.log.debug("Timeout limit reached for pythonCall", function.__name__)
                    self.__killPid(pid)

                    # HACK to avoid python bug
                    # self.wait()
                    retries = 10000
                    while os.waitpid(pid, 0) == -1 and retries > 0:
                        time.sleep(0.001)
                        retries -= 1

                    return S_ERROR('%d seconds timeout for "%s" call' % (self.timeout, function.__name__))
                elif readSeq[0] == readFD:
                    retDict = self.__readFromFD(readFD)
                    os.waitpid(pid, 0)
                    if retDict["OK"]:
                        dataStub = retDict["Value"]
                        if not dataStub:
                            return S_ERROR("Error decoding data coming from call")
                        retObj, stubLen = DEncode.decode(dataStub.encode())
                        if stubLen == len(dataStub):
                            return retObj
                        return S_ERROR("Error decoding data coming from call")
                    return retDict
            finally:
                os.close(readFD)

    def __generateSystemCommandError(self, exitStatus, message):
        """create system command error

        :param int exitStatus: exist status
        :param str message: error message
        :return: S_ERROR with additional 'Value' tuple ( existStatus, stdoutBuf, stderrBuf )
        """
        retDict = S_ERROR(message)
        retDict["Value"] = (exitStatus, self.bufferList[0][0], self.bufferList[1][0])
        return retDict

    def __readFromFile(self, fd, baseLength):
        """read from file descriptor :fd: and save it to the dedicated buffer"""
        try:
            dataString = ""
            fn = fd.fileno()

            sel = selectors.DefaultSelector()
            sel.register(fd, selectors.EVENT_READ)
            while True:
                if not sel.select(timeout=1):
                    break
                if isinstance(fn, int):
                    nB = os.read(fn, self.bufferLimit)
                else:
                    nB = fd.read(1)
                if not nB:
                    break
                try:
                    dataString += nB.decode()
                except UnicodeDecodeError as e:
                    self.log.warn("Unicode decode error in readFromFile", "(%r): %r" % (e, dataString))
                    dataString += nB.decode("utf-8", "replace")
                # break out of potential infinite loop, indicated by dataString growing beyond reason
                if len(dataString) + baseLength > self.bufferLimit:
                    self.log.error("DataString is getting too long (%s): %s " % (len(dataString), dataString[-10000:]))
                    break
        except Exception as x:
            self.log.exception("SUBPROCESS: readFromFile exception")
            try:
                self.log.error("Error reading", "type(nB) =%s" % type(nB))
                self.log.error("Error reading", "nB =%s" % str(nB))
            except Exception:
                pass
            return S_ERROR("Can not read from output: %s" % str(x))
        if len(dataString) + baseLength > self.bufferLimit:
            self.log.error("Maximum output buffer length reached")
            retDict = S_ERROR(
                "Reached maximum allowed length (%d bytes) for called " "function return value" % self.bufferLimit
            )
            retDict["Value"] = dataString
            return retDict

        return S_OK(dataString)

    def __readFromSystemCommandOutput(self, fd, bufferIndex):
        """read stdout from file descriptor :fd:"""
        retDict = self.__readFromFile(fd, len(self.bufferList[bufferIndex][0]))
        if retDict["OK"]:
            self.bufferList[bufferIndex][0] += retDict["Value"]
            if self.callback is not None:
                while self.__callLineCallback(bufferIndex):
                    pass
            return S_OK()
        else:  # buffer size limit reached killing process (see comment on __readFromFile)
            exitStatus = self.killChild()
            return self.__generateSystemCommandError(exitStatus, "%s for '%s' call" % (retDict["Message"], self.cmdSeq))

    def systemCall(self, cmdSeq, callbackFunction=None, shell=False, env=None):
        """system call (no shell) - execute :cmdSeq:"""

        if shell:
            self.log.verbose("shellCall:", cmdSeq)
        else:
            self.log.verbose("systemCall:", cmdSeq)

        self.cmdSeq = cmdSeq
        self.callback = callbackFunction
        closefd = sys.platform.find("win") != 0
        try:
            self.child = subprocess.Popen(
                self.cmdSeq,
                shell=shell,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                close_fds=closefd,
                env=env,
                universal_newlines=True,
            )
            self.childPID = self.child.pid
        except OSError as v:
            retDict = S_ERROR(repr(v))
            retDict["Value"] = (-1, "", str(v))
            return retDict
        except Exception as x:
            try:
                self.child.stdout.close()
                self.child.stderr.close()
            except Exception:
                pass
            retDict = S_ERROR(repr(x))
            retDict["Value"] = (-1, "", str(x))
            return retDict

        try:
            self.bufferList = [["", 0], ["", 0]]
            initialTime = time.time()

            exitStatus = self.__poll(self.child.pid)

            while (0, 0) == exitStatus or exitStatus is None:
                retDict = self.__readFromCommand()
                if not retDict["OK"]:
                    return retDict

                if self.timeout and time.time() - initialTime > self.timeout:
                    exitStatus = self.killChild()
                    self.__readFromCommand()
                    return self.__generateSystemCommandError(
                        exitStatus, "Timeout (%d seconds) for '%s' call" % (self.timeout, cmdSeq)
                    )
                time.sleep(0.01)
                exitStatus = self.__poll(self.child.pid)

            self.__readFromCommand()

            if exitStatus:
                exitStatus = exitStatus[1]

            if exitStatus >= 256:
                exitStatus = int(exitStatus / 256)
            return S_OK((exitStatus, self.bufferList[0][0], self.bufferList[1][0]))
        finally:
            try:
                self.child.stdout.close()
                self.child.stderr.close()
            except Exception:
                pass

    def getChildPID(self):
        """child pid getter"""
        return self.childPID

    def __readFromCommand(self):
        """read child stdout and stderr"""
        fdList = []
        for i in (self.child.stdout, self.child.stderr):
            try:
                if not i.closed:
                    fdList.append(i.fileno())
            except Exception:
                self.log.exception("SUBPROCESS: readFromCommand exception")
        readSeq = self.__selectFD(fdList, True)
        if readSeq is False:
            return S_OK()
        if self.child.stdout.fileno() in readSeq:
            retDict = self.__readFromSystemCommandOutput(self.child.stdout, 0)
            if not retDict["OK"]:
                return retDict
        if self.child.stderr.fileno() in readSeq:
            retDict = self.__readFromSystemCommandOutput(self.child.stderr, 1)
            if not retDict["OK"]:
                return retDict
        return S_OK()

    def __callLineCallback(self, bufferIndex):
        """line callback execution"""
        nextLineIndex = self.bufferList[bufferIndex][0][self.bufferList[bufferIndex][1] :].find("\n")
        if nextLineIndex > -1:
            try:
                self.callback(
                    bufferIndex,
                    self.bufferList[bufferIndex][0][
                        self.bufferList[bufferIndex][1] : self.bufferList[bufferIndex][1] + nextLineIndex
                    ],
                )
                # Each line processed is taken out of the buffer to prevent the limit from killing us
                nL = self.bufferList[bufferIndex][1] + nextLineIndex + 1
                self.bufferList[bufferIndex][0] = self.bufferList[bufferIndex][0][nL:]
                self.bufferList[bufferIndex][1] = 0
            except Exception:
                self.log.exception("Exception while calling callback function", "%s" % self.callback.__name__)
                self.log.showStack()
                return False

            return True
        return False


def systemCall(timeout, cmdSeq, callbackFunction=None, env=None, bufferLimit=52428800):
    """
    Use SubprocessExecutor class to execute cmdSeq (it can be a string or a sequence)
    with a timeout wrapper, it is executed directly without calling a shell
    """
    if timeout > 0 and USE_WATCHDOG:
        spObject = Subprocess(timeout=timeout, bufferLimit=bufferLimit)
        sysCall = Watchdog(
            spObject.systemCall,
            args=(cmdSeq,),
            kwargs={"callbackFunction": callbackFunction, "env": env, "shell": False},
        )
        spObject.log.verbose("Subprocess Watchdog timeout set to %d" % timeout)
        result = sysCall(timeout + 1)
    else:
        spObject = Subprocess(timeout, bufferLimit=bufferLimit)
        result = spObject.systemCall(cmdSeq, callbackFunction=callbackFunction, env=env, shell=False)
    return result


def shellCall(timeout, cmdSeq, callbackFunction=None, env=None, bufferLimit=52428800):
    """
    Use SubprocessExecutor class to execute cmdSeq (it can be a string or a sequence)
    with a timeout wrapper, cmdSeq it is invoque by /bin/sh
    """
    if timeout > 0 and USE_WATCHDOG:
        spObject = Subprocess(timeout=timeout, bufferLimit=bufferLimit)
        shCall = Watchdog(
            spObject.systemCall,
            args=(cmdSeq,),
            kwargs={"callbackFunction": callbackFunction, "env": env, "shell": True},
        )
        spObject.log.verbose("Subprocess Watchdog timeout set to %d" % timeout)
        result = shCall(timeout + 1)
    else:
        spObject = Subprocess(timeout, bufferLimit=bufferLimit)
        result = spObject.systemCall(cmdSeq, callbackFunction=callbackFunction, env=env, shell=True)
    return result


def pythonCall(timeout, function, *stArgs, **stKeyArgs):
    """
    Use SubprocessExecutor class to execute function with provided arguments,
    with a timeout wrapper.
    """
    if timeout > 0 and USE_WATCHDOG:
        spObject = Subprocess(timeout=timeout)
        pyCall = Watchdog(spObject.pythonCall, args=(function,) + stArgs, kwargs=stKeyArgs)
        spObject.log.verbose("Subprocess Watchdog timeout set to %d" % timeout)
        result = pyCall(timeout + 1)
    else:
        spObject = Subprocess(timeout)
        result = spObject.pythonCall(function, *stArgs, **stKeyArgs)
    return result


def getChildrenPIDs(ppid, foreachFunc=None):
    """
    Get all children recursively for a given ppid.
     Optional foreachFunc will be executed for each children pid
    """
    cpids = psutil.Process(ppid).children(recursive=True)
    pids = []
    for proc in cpids:
        pids.append(proc.pid)
        if foreachFunc:
            foreachFunc(proc.pid)
    return pids



""" VOMSSecurityManager class implements access permissions based on VOMS roles
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import datetime

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getAllGroups, getGroupOption
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.SecurityManagerBase import (
    SecurityManagerBase,
    _readMethods,
    _writeMethods,
)
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult


class VOMSSecurityManager(SecurityManagerBase):
    """This class implements a 3-level POSIX permission, wrapping up
    the DIRAC group into VOMS roles
    """

    def __init__(self, database=False):
        super(VOMSSecurityManager, self).__init__(database=database)

        # voms role : [dirac groups that have it]
        self.vomsRoles = {}
        # dirac group : voms role it has
        self.diracGroups = {}

        # Lifetime of the info in the two dictionaries
        self.CACHE_TIME = datetime.timedelta(seconds=600)
        self.__buildRolesAndGroups()

    def __buildRolesAndGroups(self):
        """Rebuild the cache dictionary for VOMS roles and DIRAC Groups"""

        self.lastBuild = datetime.datetime.now()

        allGroups = getAllGroups()

        for grpName in allGroups:
            vomsRole = getGroupOption(grpName, "VOMSRole")
            if vomsRole:
                self.diracGroups[grpName] = vomsRole
                self.vomsRoles.setdefault(vomsRole, []).append(grpName)

    def __getVomsRole(self, grpName):
        """Returns the VOMS role of a given DIRAC group

        :param grpName:

        :returns: VOMS role, or None
        """
        if (datetime.datetime.now() - self.lastBuild) > self.CACHE_TIME:
            self.__buildRolesAndGroups()

        return self.diracGroups.get(grpName)

    def __getDiracGroups(self, vomsRole):
        """Returns all the DIRAC groups that have a given VOMS role

        :param vomsRole:

        :returns: list of groups, empty if not exist
        """

        if (datetime.datetime.now - self.lastBuild) > self.CACHE_TIME:
            self.__buildRolesAndGroups()

        return self.vomsRoles.get(vomsRole, [])

    def __shareVomsRole(self, grpName, otherGrpName):
        """Returns True if the two DIRAC groups have the same VOMS role"""

        vomsGrp = self.__getVomsRole(grpName)
        vomsOtherGrp = self.__getVomsRole(otherGrpName)
        # The voms group cannot be None
        return vomsGrp and vomsOtherGrp and (vomsGrp == vomsOtherGrp)

    def __isNotExistError(self, errorMsg):
        """Returns true if the errorMsg means that the file/directory does not exist"""

        for possibleMsg in ["not exist", "not found", "No such file or directory"]:
            if possibleMsg in errorMsg:
                return True

        return False

    def __getFilePermission(self, path, credDict, noExistStrategy=None):
        """Checks POSIX permission for a file using the VOMS roles.
        That is, if the owner group of the file shares the same vomsRole as the requesting user,
        we check the permission as if the request was done with the real owner group.

        :param path : file path (string)
        :param credDict : credential of the user
        :param noExistStrategy : If the directory does not exist, we can
                                 * True : allow the access
                                 * False : forbid the access
                                 * None : return the error as is

        :returns S_OK structure with a dictionary ( Read/Write/Execute : True/False)
        """

        if not path:
            return S_ERROR("Empty path")

        # We check what is the group stored in the DB for the given path
        res = returnSingleResult(self.db.fileManager.getFileMetadata([path]))
        if not res["OK"]:
            # If the error is not due to the directory not existing, we return
            if not self.__isNotExistError(res["Message"]):
                return res

            # From now on, we know that the error is due to the file not existing

            # If we have no strategy regarding non existing files, then just return the error
            if noExistStrategy is None:
                return res

            # Finally, follow the strategy
            return S_OK(dict.fromkeys(["Read", "Write", "Execute"], noExistStrategy))

        # ===========================================================================
        # # This does not seem necessary since we add the OwnerGroup in the query behind the scene
        # origGrp = 'unknown'
        # res = self.db.ugManager.getGroupName( res['Value']['GID'] )
        # if res['OK']:
        #   origGrp = res['Value']
        # ===========================================================================

        origGrp = res["Value"].get("OwnerGroup", "unknown")

        # If the two group share the same voms role, we do the query like if we were
        # the group stored in the DB
        if self.__shareVomsRole(credDict.get("group", "anon"), origGrp):
            credDict = {"username": credDict.get("username", "anon"), "group": origGrp}

        return returnSingleResult(self.db.fileManager.getPathPermissions([path], credDict))

    def __testPermissionOnFile(self, paths, permission, credDict, noExistStrategy=None):
        """Tests a permission on a list of files

        :param path : list/dict of file paths
        :param permission : Read/Write/Execute string
        :param credDict : credential of the user
        :param noExistStrategy : If the directory does not exist, we can
                                 * True : allow the access
                                 * False : forbid the access
                                 * None : return the error as is

        :returns: Successful dictionary with True of False, and Failed.
        """

        successful = {}
        failed = {}

        for filename in paths:
            res = self.__getFilePermission(filename, credDict, noExistStrategy=noExistStrategy)
            if not res["OK"]:
                failed[filename] = res["Message"]
            else:
                successful[filename] = res["Value"].get(permission, False)

        return S_OK({"Successful": successful, "Failed": failed})

    def __getDirectoryPermission(self, path, credDict, recursive=True, noExistStrategy=None):
        """Checks POSIX permission for a directory using the VOMS roles.
        That is, if the owner group of the directory share the same vomsRole as the requesting user,
        we check the permission as if the request was done with the real owner group.

        :param path : directory path (string)
        :param credDict : credential of the user
        :param recursive : if that directory does not exist, checks the parent one
        :param noExistStrategy : If the directory does not exist, we can
                                 * True : allow the access
                                 * False : forbid the access
                                 * None : return the error as is

               noExistStrategy makes sense only if recursive is False

        :returns S_OK structure with a dictionary ( Read/Write/Execute : True/False)
        """

        if not path:
            return S_ERROR("Empty path")

        # We check what is the group stored in the DB for the given path
        res = self.db.dtree.getDirectoryParameters(path)
        if not res["OK"]:
            # If the error is not due to the directory not existing, we return

            if not self.__isNotExistError(res["Message"]):
                return res

            # Very special case to allow creation of very first entry
            if path == "/":
                return S_OK({"Read": True, "Write": True, "Execute": True})

            # From now on, we know that the error is due to the directory not existing

            # If recursive, we try the parent directory
            if recursive:
                return self.__getDirectoryPermission(
                    os.path.dirname(path), credDict, recursive=recursive, noExistStrategy=noExistStrategy
                )
            # From now on, we know we don't run recursive

            # If we have no strategy regarding non existing directories, then just return the error
            if noExistStrategy is None:
                return res

            # Finally, follow the strategy
            return S_OK(dict.fromkeys(["Read", "Write", "Execute"], noExistStrategy))

        # The directory exists.
        origGrp = res["Value"]["OwnerGroup"]

        # If the two group share the same voms role, we do the query like if we were
        # the group stored in the DB
        if self.__shareVomsRole(credDict.get("group", "anon"), origGrp):
            credDict = {"username": credDict.get("username", "anon"), "group": origGrp}

        return self.db.dtree.getDirectoryPermissions(path, credDict)

    def __testPermissionOnDirectory(self, paths, permission, credDict, recursive=True, noExistStrategy=None):
        """Tests a permission on a list of directories

        :param path : list/dict of directory paths
        :param permission : Read/Write/Execute string
        :param credDict : credential of the user
        :param recursive : if that directory does not exist, checks the parent one
        :param noExistStrategy : If the directory does not exist, we can
                                 * True : allow the access
                                 * False : forbid the access
                                 * None : return the error as is

               noExistStrategy makes sense only if recursive is False

        :returns: Successful dictionary with True of False, and Failed.
        """

        successful = {}
        failed = {}

        for dirName in paths:
            res = self.__getDirectoryPermission(dirName, credDict, recursive=recursive, noExistStrategy=noExistStrategy)
            if not res["OK"]:
                failed[dirName] = res["Message"]
            else:
                successful[dirName] = res["Value"].get(permission, False)

        return S_OK({"Successful": successful, "Failed": failed})

    def __testPermissionOnParentDirectory(self, paths, permission, credDict, recursive=True, noExistStrategy=None):
        """Tests a permission on the parents of a list of directories

        :param path : directory path (string)
        :param permission : Read/Write/Execute string
        :param credDict : credential of the user
        :param recursive : if that directory does not exist, checks the parent one
        :param noExistStrategy : If the directory does not exist, we can
                                 * True : allow the access
                                 * False : forbid the access
                                 * None : return the error as is

               noExistStrategy makes sense only if recursive is False

        :returns: Successful dictionary with True of False, and Failed.
        """
        parentDirs = {}
        for path in paths:
            parentDirs.setdefault(os.path.dirname(path), []).append(path)

        res = self.__testPermissionOnDirectory(
            parentDirs, permission, credDict, recursive=recursive, noExistStrategy=noExistStrategy
        )

        if not res["OK"]:
            return res

        failed = res["Value"]["Failed"]
        successful = {}

        parentAllowed = res["Value"]["Successful"]

        for parentName in parentAllowed:
            isParentAllowed = parentAllowed[parentName]
            for path in parentDirs[parentName]:
                successful[path] = isParentAllowed

        return S_OK({"Successful": successful, "Failed": failed})

    def __getFileOrDirectoryPermission(self, path, credDict, recursive=False, noExistStrategy=None):
        """Checks POSIX permission for a directory or file using the VOMS roles.
        That is, if the owner group of the directory or file shares the same vomsRole as the requesting user,
        we check the permission as if the request was done with the real owner group.

        We first consider the path as a file, and if it does not exist, we consider it as a directory.

        :param path : directory or file path (string)
        :param credDict : credential of the user
        :param recursive : if that directory does not exist, checks the parent one
        :param noExistStrategy : If the directory does not exist, we can
                                 * True : allow the access
                                 * False : forbid the access
                                 * None : return the error as is

               noExistStrategy makes sense only if recursive is False

        :returns S_OK structure with a dictionary ( Read/Write/Execute : True/False)
        """
        # First consider it as File
        # We want to know whether the file does not exist, so we force noExistStrategy to None
        res = self.__getFilePermission(path, credDict, noExistStrategy=None)
        if not res["OK"]:
            # If the error is not due to the directory not existing, we return
            if not self.__isNotExistError(res["Message"]):
                return res

            # From now on, we know that the error is due to the File not existing
            # We Try then the directory method, since path can be a directory
            # The noExistStrategy will be applied by __getDirectoryPermission, so we don't need to do it ourselves
            res = self.__getDirectoryPermission(path, credDict, recursive=recursive, noExistStrategy=noExistStrategy)

        return res

    def __testPermissionOnFileOrDirectory(self, paths, permission, credDict, recursive=False, noExistStrategy=None):
        """Tests a permission on a list of files or directories.

        :param path : list/dict of directory or files paths
        :param permission : Read/Write/Execute string
        :param credDict : credential of the user
        :param recursive : if that directory does not exist, checks the parent one
        :param noExistStrategy : If the directory does not exist, we can
                                 * True : allow the access
                                 * False : forbid the access
                                 * None : return the error as is

               noExistStrategy makes sense only if recursive is False

        :returns: Successful dictionary with True of False, and Failed.
        """

        successful = {}
        failed = {}

        for path in paths:
            res = self.__getFileOrDirectoryPermission(
                path, credDict, recursive=recursive, noExistStrategy=noExistStrategy
            )
            if not res["OK"]:
                failed[path] = res["Message"]
            else:
                successful[path] = res["Value"].get(permission, False)

        return S_OK({"Successful": successful, "Failed": failed})

    def __policyRemoveDirectory(self, paths, credDict):
        """Tests whether the remove operation on directories
        is permitted.
        Removal of non existing directory is always allowed.
        For existing directories, we must have the write permission
        on the parent

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """

        successful = {}

        # We allow removal of all the non existing directories
        res = self.db.dtree.exists(paths)
        if not res["OK"]:
            return res

        nonExistingDirectories = set(
            path for path in res["Value"]["Successful"] if not res["Value"]["Successful"][path]
        )

        existingDirs = set(paths) - set(nonExistingDirectories)
        for dirName in nonExistingDirectories:
            successful[dirName] = True

        res = self.__testPermissionOnParentDirectory(existingDirs, "Write", credDict, recursive=False)
        if not res["OK"]:
            return res

        failed = res["Value"]["Failed"]
        successful.update(res["Value"]["Successful"])

        return S_OK({"Successful": successful, "Failed": failed})

    def __policyRemoveFile(self, paths, credDict):
        """Tests whether the remove operation on files
        is permitted.
        Removal of non existing file is always allowed.
        For existing files, we must have the write permission
        on the parent

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """

        successful = {}

        # We allow removal of all the non existing files
        res = self.db.fileManager.exists(paths)
        if not res["OK"]:
            return res

        nonExistingFiles = set(path for path in res["Value"]["Successful"] if not res["Value"]["Successful"][path])

        existingFiles = set(paths) - set(nonExistingFiles)
        for dirName in nonExistingFiles:
            successful[dirName] = True

        res = self.__testPermissionOnParentDirectory(existingFiles, "Write", credDict, recursive=False)
        if not res["OK"]:
            return res

        failed = res["Value"]["Failed"]
        successful.update(res["Value"]["Successful"])

        return S_OK({"Successful": successful, "Failed": failed})

    def __policyListDirectory(self, paths, credDict):
        """Test Read permission on the directory.
        If the directory does not exist, we do not allow.

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """

        return self.__testPermissionOnDirectory(paths, "Read", credDict, recursive=True)

    def __policyReadForFileAndDirectory(self, paths, credDict):
        """Testing the read bit on the parent directory,
        be it a file or a directory.
        So it reads permissions from a directory

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """

        return self.__testPermissionOnParentDirectory(paths, "Read", credDict, recursive=True)

    def __policyWriteForFileAndDirectory(self, paths, credDict):
        """Testing the read bit on the parent directory (recursively),
        be it a file or a directory.
        So it reads permissions from a directory

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """

        return self.__testPermissionOnParentDirectory(paths, "Write", credDict, recursive=True)

    def __policyReadForReplica(self, paths, credDict):
        """Test Read permission on the file associated to the replica.
        If the file does not exist, we allow.

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """
        return self.__testPermissionOnFile(paths, "Read", credDict, noExistStrategy=True)

    def __policyWriteForReplica(self, paths, credDict):
        """Test Write permission on the file associated to the replica.
        If the file does not exist, we allow.

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """
        return self.__testPermissionOnFile(paths, "Write", credDict, noExistStrategy=True)

    def __policyWriteOnFile(self, paths, credDict):
        """Test Write permission on the file.
        If the file does not exist, we allow.

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """
        return self.__testPermissionOnFile(paths, "Write", credDict, noExistStrategy=True)

    def __policyChangePathMode(self, paths, credDict):
        """Test Write permission on the directory/file.
        If the directory/file does not exist, we allow.

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """

        return self.__testPermissionOnFileOrDirectory(paths, "Write", credDict, recursive=False, noExistStrategy=True)

    def __policyDeny(self, paths, credDict):
        """Denies the access to all the paths given

        :param paths: list/dict of path
        :param credDict: credential of the user
        :returns: Successful with True of False, and Failed.
        """

        return S_OK({"Successful": dict.fromkeys(paths, False), "Failed": {}})

    def hasAccess(self, opType, paths, credDict):
        """Checks whether a given operation on given paths is permitted

        :param opType: name of the operation (the FileCatalog methods in fact...)
        :param paths: list/dictionary of path on which we want to apply the operation
        :param credDict: credential of the users (with at least username, group and properties)

        :returns: Successful dict with True or False, and Failed dict. In fact, it is not neccesarily
                a boolean, rather an int (binary operation results)
        """

        # Check if admin access is granted first
        result = self.hasAdminAccess(credDict)
        if not result["OK"]:
            return result

        if result["Value"]:
            # We are admin, allow everything
            return S_OK({"Successful": dict.fromkeys(paths, True), "Failed": {}})

        if opType not in _readMethods + _writeMethods:
            return S_ERROR("Operation type not known %s" % opType)

        if self.db.globalReadAccess and (opType in _readMethods):
            return S_OK({"Successful": dict.fromkeys(paths, True), "Failed": {}})

        policyToExecute = None

        if opType == "removeDirectory":
            policyToExecute = self.__policyRemoveDirectory

        elif opType in ["createDirectory", "addFile"]:
            policyToExecute = self.__policyWriteForFileAndDirectory

        elif opType == "removeFile":
            policyToExecute = self.__policyRemoveFile

        elif opType in [
            "addFileAncestors",
            "setFileStatus",
            "addReplica",
            "removeReplica",
            "setReplicaStatus",
            "setReplicaHost",
        ]:
            policyToExecute = self.__policyWriteOnFile

        elif opType == "listDirectory":
            policyToExecute = self.__policyListDirectory

        elif opType in [
            "isDirectory",
            "getDirectoryReplicas",
            "getDirectoryMetadata",
            "getDirectorySize",
            "isFile",
            "getFileSize",
            "getFileMetadata",
            "exists",
            "getFileAncestors",
            "getFileDescendents",
        ]:
            policyToExecute = self.__policyReadForFileAndDirectory

        elif opType in ["getReplicas", "getReplicaStatus"]:
            policyToExecute = self.__policyReadForReplica

        # Only admin can do that, and if we are here, we are not admin
        elif opType in ["changePathOwner", "changePathGroup"]:
            policyToExecute = self.__policyDeny

        elif opType == "changePathMode":
            policyToExecute = self.__policyChangePathMode

        if not policyToExecute:
            return S_ERROR("No policy matching operation %s" % opType)

        res = policyToExecute(paths, credDict)

        return res

    def getPathPermissions(self, paths, credDict):
        """This method is meant to disappear, hopefully soon,
        but as long as we have clients from versions < v6r14,
        we need a getPathPermissions method. Since it does not make
        sense with that kind of fine grain policy, we return what used to
        be returned...
        """
        return super(VOMSSecurityManager, self).getPathPermissions(paths, credDict)
